import re
from datetime import date
from typing import Dict, List, Union

from lxml.html import HtmlElement

from juriscraper.pacer.reports import BaseReport

from ..lib.log_tools import make_default_logger
from ..lib.string_utils import convert_date_string, force_unicode, harmonize
from ..lib.utils import clean_court_object
from .docket_report import BaseDocketReport
from .utils import get_pacer_case_id_from_doc1_url

logger = make_default_logger()


class ClaimsActivity(BaseDocketReport, BaseReport):
    """Query and parse the claims activity."""

    PATH = "cgi-bin/ClaimsActRpt.pl"
    CACHE_ATTRS = ["claims", "metadata"]
    ERROR_STRINGS = BaseReport.ERROR_STRINGS + [
        "No claims entries found",
        "Security violation: You do not have access rights to this program",
    ]

    def __init__(self, court_id, pacer_session=None):
        BaseDocketReport.__init__(self, court_id)
        BaseReport.__init__(self, court_id, pacer_session)

        self._clear_caches()
        self._metadata = None
        self._claims = None
        assert court_id.endswith(
            "b"
        ), "Unable to create object. Must use bankruptcy court abbreviation."

    @property
    def data(self):
        """Get all the data back from this endpoint."""
        if self.is_valid is False:
            return {}

        data = self.metadata.copy()
        return data

    @property
    def metadata(self):
        """
        Each claim table looks like this:

        --------------------+------------+---------------------
        Case and claim data | Claim data |  Case and claim data
        -------------------------------------------------------
        Claim data
        -------------------------------------------------------
        Claim data (no always present)
        -------------------------------------------------------

        So first we parse each colum of the first row, then second and third
        row, finding the fields for the case and the claim.
        """

        if self.is_valid is False:
            return {}

        if self._metadata is not None:
            return self._metadata

        data = []

        field_mappings_case = {
            "title": "case_name",
            "case": "docket_number",
            "office": "office",
            "chapter": "chapter",
            "judge": "assigned_to_str",
            "trustee": "trustee_str",
            "last_date_to_file_claims": "last_date_to_file_claims",
            "last_date_to_file_govt": "last_date_to_file_govt",
        }
        field_mappings_claim = {
            "amends_no": "amends_no",
            "amount_claimed": "amount_claimed",
            "amount_allowed": "amount_allowed",
            "secured_claimed": "secured_claimed",
            "priority_claimed": "priority_claimed",
            "creditor_name_address": "creditor_name_address",
            "description": "description",
            "creditor_id": "creditor_id",
            "filed": "date_filed",
            "entered": "date_entered",
            "entered_by": "entered_by",
            "filed_by": "filed_by",
            "status": "status",
            "remarks": "remarks",
        }

        claim_tables = self.tree.xpath('//table[@class="complexReport"]')
        for claim_table in claim_tables:
            meta_data = {
                "court_id": self.court_id,
            }
            claim = {}

            label_nodes = claim_table.xpath("(.//tr)[1]/td[1]/text()")
            # Parse case and claim data from row 1 column 1.
            for label_node in label_nodes:
                meta_data.update(
                    self._get_label_value_pair_from_string(
                        label_node, field_mappings_case
                    )
                )
                claim.update(
                    self._get_label_value_pair_from_string(
                        label_node, field_mappings_claim
                    )
                )

            self._parse_case_name_and_docket_number(meta_data, claim_table)

            claim_anchor_nodes = claim_table.xpath("(.//tr)[1]/td[1]/b/a")
            claim.update(self._get_claim_data_from_anchors(claim_anchor_nodes))

            # Parse claim data from row 1 column 2.
            label_nodes = claim_table.xpath("(.//tr)[1]/td[2]/text()")
            for label_node in label_nodes:
                claim.update(
                    self._get_label_value_pair_from_string(
                        label_node, field_mappings_claim
                    )
                )

            # Parse case and claim data from row 1 column 3.
            label_nodes = claim_table.xpath("(.//tr)[1]/td[3]/text()")
            for label_node in label_nodes:
                meta_data.update(
                    self._get_label_value_pair_from_string(
                        label_node, field_mappings_case
                    )
                )
                claim.update(
                    self._get_label_value_pair_from_string(
                        label_node, field_mappings_claim
                    )
                )

            # Parse claim data from row 2.
            label_nodes = claim_table.xpath("(.//tr)[2]/td[1]/text()")
            for label_node in label_nodes:
                claim.update(
                    self._get_label_value_pair_from_string(
                        label_node, field_mappings_claim
                    )
                )

            # Parse claim data from row 3.
            label_nodes = claim_table.xpath("(.//tr)[3]/td[1]/text()")
            for label_node in label_nodes:
                claim.update(
                    self._get_label_value_pair_from_string(
                        label_node, field_mappings_claim
                    )
                )

            meta_data["claim"] = claim
            data.append(meta_data)

            # Fill non present values in the report in order to standarize
            # fields in the report.
            self._fill_non_present_values_as_empty(
                field_mappings_case, meta_data
            )
            self._fill_non_present_values_as_empty(
                field_mappings_claim, meta_data["claim"]
            )

        data = clean_court_object(data)
        self._metadata = data
        return data

    def _fill_non_present_values_as_empty(
        self,
        field_mappings: Dict[str, str],
        meta_data: Dict[str, Union[str, date, None, List[Dict[str, str]]]],
    ) -> None:
        """Fill claim missing values in meta_data with empty values if they are
        not present.

        :param field_mappings: A dictionary of mappings field names.
        :param meta_data: A dictionary containing the metadata for the claim.
        :return: None, meta_data is modified in place.
        """

        for key, value in field_mappings.items():
            if value not in meta_data:
                if value.startswith("date_") or value.startswith("last_date_"):
                    meta_data[value] = None
                else:
                    meta_data[value] = ""

    def _parse_case_name_and_docket_number(
        self,
        meta_data: Dict[str, Union[str, date, None, List[Dict[str, str]]]],
        claim_table: HtmlElement,
    ) -> None:
        """Parses the case name and docket number from the provided meta_data
        dictionary and claim_table HTML element.

        :param meta_data: A dictionary containing the case metadata.
        :param claim_table: The lxml.HtmlElement claim_table
        :return: None, the meta_data dict is updated in place.
        """
        meta_data["case_name"] = harmonize(meta_data["case_name"])

        docket_number = meta_data.get("docket_number")
        if docket_number:
            meta_data["docket_number"] = self._clean_docket_number(
                docket_number
            )
        else:
            docket_number = claim_table.xpath("(.//tr)[1]/td[1]/a//text()")[0]
            meta_data["docket_number"] = self._clean_docket_number(
                docket_number
            )

    def _clean_docket_number(self, docket_number: str) -> str:
        """Strip the judge initials off docket numbers, and do any other
        similar cleanup.

        :param: docket_number: A string docket number to clean.
        :return: The docket_number cleaned.
        """

        # Uses both b/c sometimes the bankr. cases have a dist-style docket
        # number.
        regexes = [
            self.docket_number_dist_regex,
            self.docket_number_bankr_regex,
        ]
        for regex in regexes:
            match = regex.search(docket_number)
            if match:
                return match.group(1)

        return docket_number

    def _get_claim_data_from_anchors(
        self, anchor_nodes: List[HtmlElement]
    ) -> Dict[str, Union[str, List[Dict[str, str]]]]:
        """Retrieves claim data from a list of HTML anchor elements.

        :param anchor_nodes: A list of HTML anchor elements that contain
        information about the claim.

        :return: A dict with the claim data and claim attachments.
        """

        claim = {}
        attachments = []
        for i, anchor in enumerate(anchor_nodes):
            document_url = (
                anchor.xpath("./@href")[0] if anchor is not None else None
            )
            if i == 0:
                claim["pacer_case_id"] = get_pacer_case_id_from_doc1_url(
                    document_url
                )
                claim["claim_id"] = self.get_pacer_claim_id_from_claim_url(
                    document_url
                )
                claim["claim_number"] = self.get_claim_number_from_claim_url(
                    document_url
                )
            else:
                claim_att = {}
                claim_att["pacer_case_id"] = get_pacer_case_id_from_doc1_url(
                    document_url
                )
                claim_att["claim_id"] = self.get_pacer_claim_id_from_claim_url(
                    document_url
                )
                claim_att[
                    "claim_number"
                ] = self.get_claim_number_from_claim_url(document_url)
                claim_att["short_description"] = anchor.text_content()
                claim_att[
                    "claim_doc_seq"
                ] = self.get_claim_doc_seq_from_claim_url(document_url)

                attachments.append(claim_att)

        claim["attachments"] = attachments
        return claim

    @staticmethod
    def get_pacer_claim_id_from_claim_url(url: str) -> Union[str, None]:
        """Extract the caseid from the doc1 URL."""
        match = re.search(r"claim_id=(\d+)", url)
        if match:
            return match.group(1)
        else:
            return None

    @staticmethod
    def get_claim_number_from_claim_url(url: str) -> Union[str, None]:
        """Extract the caseid from the doc1 URL."""
        match = re.search(r"claim_num=(\d+-\d+)", url)
        if match:
            return match.group(1)
        else:
            return None

    @staticmethod
    def get_claim_doc_seq_from_claim_url(url: str) -> Union[str, None]:
        """Extract the caseid from the doc1 URL."""
        match = re.search(r"claim_doc_seq=(\d+)", url)
        if match:
            return match.group(1)
        else:
            return None

    @staticmethod
    def _get_label_value_pair_from_string(
        string: str, field_mappings: Dict[str, str]
    ) -> Dict[str, Union[str, date, None]]:
        """Get the field name and value from a string.

        e.g: Case: 3:23-bk-10722
        :param string: The string where to extract the label and value.
        :param field_mappings: A dict mapping PACER labels to Juriscraper ones.
        :return a dict with a k-v mapping between a label and its value.
        """

        splited_string = string.split(":", 1)

        if len(splited_string) <= 1:
            return {}

        label = splited_string[0]
        value = splited_string[1]

        label = (
            label.strip()
            .lower()
            .replace(" ", "_")
            .replace("/", "_")
            .replace("\xa0", "_")  # Non-breaking space
            .replace("(", "")
            .replace(")", "")
            .rstrip(":")
        )
        label = field_mappings.get(label, None)

        if not label:
            return {}

        value = value.strip()
        value = value.lstrip(":").strip()
        if label.startswith("date_") or label.startswith("last_date_"):
            if value:
                data = {label: convert_date_string(value)}
            else:
                data = {label: None}
        else:
            data = {label: force_unicode(value)}
        return data

    def query(
        self,
        pacer_case_id: str,
        docket_number: str,
        creditor_name: str,
        date_start: date = None,
        date_end: date = None,
    ):
        """Query the claims activity and return the results.

        :param pacer_case_id: The internal PACER ID for the case.
        :param docket_number: A docket number to look up. Something like
        2:17-bk-39239.
        :param creditor_name: The creditor name to look up.
        :param date_start: The earliest claim entry you wish to have. Default
        is 1960-01-01 (before anything was put in PACER)
        :param date_end: The latest claim entry you with to have. Default is
        tomorrow.

        :return: request response object
        """
        assert (
            self.session is not None
        ), "session attribute of ClaimsActivity cannot be None."

        params = {
            "all_case_ids": pacer_case_id,
            # Whether to filter by filed date or entered date. Parser only
            # currently supports filed date.
            "sort1": "case number",
            "sort2": "claim number",
            # This field seems to be required while 'all_case_ids' is not.
            "case_num": docket_number,
            "creditor": creditor_name,
        }
        if date_start:
            params["entered_from"] = date_start.strftime("%m/%d/%Y")
        else:
            # Set the default
            params["entered_from"] = "1/1/1960"
        if date_end:
            params["entered_to"] = date_end.strftime("%m/%d/%Y")

        logger.info(
            "Querying claims activy for case ID '%s' in court '%s' "
            "with params %s",
            pacer_case_id,
            self.court_id,
            params,
        )

        if self.court_id == "insb":
            # The POST param wildcard 1-L_1_0-1 doesn't work for insb.
            # It throws a 500 error. 1-L_945_0-1 doesn't work either.
            # The only approach that seems to work is using one of the params
            # generated in cgi-bin/ClaimsActRpt.pl
            # Unfortunately this might stop working since it seems to expire
            # or might be linked to the PACER session.
            post_param = "124892645626785-L_945_0-1"
        else:
            post_param = "1-L_1_0-1"

        self.response = self.session.post(
            f"{self.url}?{post_param}", data=params
        )
        self.parse()



# Scraper for Texas 9th Court of Appeals
# CourtID: texapp9
# Court Short Name: TX
# Author: Andrei Chelaru
# Reviewer: mlr
# Date: 2014-07-10


from juriscraper.opinions.united_states.state import tex


class Site(tex.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court_name = "capp_9"
        self.checkbox = 10



from juriscraper.opinions.united_states.state import nj


class Site(nj.Site):
    """Human web interface: http://www.judiciary.state.nj.us/attorneys/opinions.html#Unpublished_Appellate"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.case_types = ["Published_Appellate", "Unpublished_Appellate"]



__all__ = [
    "acca_memorandum",
    "acca_p",
    "acca_summary",
    "afcca",
    "ag",
    "armfor",
    "cavc",
    "cgcca",
    "cit",
    "nmcca",
    "tax",
    "uscfc",
    "uscfc_u",
    "uscfc_vaccine",
    "uscfc_vaccine_u",
]



import json
from pprint import pprint

import requests
from lxml import etree, html


def get_court_names():
    response = requests.get("https://www.courtlistener.com/api/jurisdictions/")
    tree = html.fromstring(response.text)

    data = dict()

    for tr in tree.xpath("//tr"):
        pre = tr.xpath("./td[5]/a/text()")
        name = tr.xpath("./td[1]/text()")
        if pre and name:
            data[pre[0].strip()] = name[0].strip()

    with open("./court_names.json", "w") as f:
        json.dump(data, f)


def get_fdsys_court_names():
    response = requests.get(
        "https://www.gpo.gov/smap/fdsys/sitemap_2014/2014_USCOURTS_sitemap.xml",
        stream=True,
    )
    response.raw.decode_content = True
    tree = etree.parse(response.raw)
    data = dict()

    for url in tree.xpath(
        "//m:loc/text()",
        namespaces={
            "m": "http://www.sitemaps.org/schemas/sitemap/0.9",
            "xlink": "http://www.w3.org/1999/xlink",
        },
    ):
        pre = url.split("-")[1]
        # if pre not in data and url:
        data[pre] = url

    with open("./fdsys_court_names_2014.json", "w") as f:
        json.dump(data, f)


if __name__ == "__main__":
    # get_court_names()
    get_fdsys_court_names()



import calendar
import re
import string
from datetime import timedelta

import geonamescache
from dateutil import parser

# For use in titlecase
BIG = (
    "3D|AFL|AKA|A/K/A|BMG|CBS|CDC|CDT|CEO|CIO|CNMI|D/B/A|DOJ|DVA|EFF|"
    "FCC|FTC|HSBC|IBM|II|III|IV|JJ|LLC|LLP|MCI|MJL|MSPB|ND|NLRB|PTO|"
    "SD|UPS|RSS|SEC|UMG|US|USA|USC|USPS|WTO"
)
SMALL = r"a|an|and|as|at|but|by|en|for|if|in|is|of|on|or|the|to|v\.?|via|vs\.?"
NUMS = "0123456789"
PUNCT = r"""!"#$¢%&'‘()*+,\-./:;?@[\\\]_—`{|}~"""
WEIRD_CHARS = r"¼½¾§ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÑÒÓÔÕÖØÙÚÛÜßàáâãäåæçèéêëìíîïñòóôœõöøùúûüÿ"
BIG_WORDS = re.compile(rf"^({BIG})[{PUNCT}]?$", re.I | re.U)
SMALL_WORDS = re.compile(r"^(%s)$" % SMALL, re.I)
SMALL_WORD_INLINE = re.compile(r"(^|\s)(%s)(\s|$)" % SMALL, re.I | re.U)
INLINE_PERIOD = re.compile(r"[a-z][.][a-z]", re.I)
INLINE_SLASH = re.compile(r"[a-z][/][a-z]", re.I)
INLINE_AMPERSAND = re.compile(r"([a-z][&][a-z])(.*)", re.I)
UC_ELSEWHERE = re.compile(r"[%s]*?[a-zA-Z]+[A-Z]+?" % PUNCT, re.U)
CAPFIRST = re.compile(r"^[%s]*?([A-Za-z])" % PUNCT)
SMALL_FIRST = re.compile(rf"^([{PUNCT}]*)({SMALL})\b", re.I | re.U)
SMALL_LAST = re.compile(rf"\b({SMALL})[{PUNCT}]?$", re.I | re.U)
SUBPHRASE = re.compile(r"([:;?!][ ])(%s)" % SMALL)
APOS_SECOND = re.compile(r"^[dol]{1}['‘]{1}[a-z]+$", re.I)
ALL_CAPS = re.compile(rf"^[A-Z\s{PUNCT}{WEIRD_CHARS}{NUMS}]+$")
UC_INITIALS = re.compile(r"^(?:[A-Z]{1}\.{1}|[A-Z]{1}\.{1}[A-Z]{1})+,?$")
MAC_MC = re.compile(r"^([Mm]a?c)(\w+.*)")


def titlecase(text, DEBUG=False):
    """Titlecases input text

    This filter changes all words to Title Caps, and attempts to be clever
    about *un*capitalizing SMALL words like a/an/the in the input.

    The list of "SMALL words" which are not capped comes from
    the New York Times Manual of Style, plus 'vs' and 'v'.

    This will fail if multiple sentences are provided as input and if the
    first word of a sentence is a SMALL_WORD.

    List of "BIG words" grows over time as entries are needed.
    """
    text_sans_small_words = re.sub(SMALL_WORD_INLINE, "", text)
    if text_sans_small_words.isupper():
        # if, after removing small words, the entire string is uppercase,
        # we lowercase it
        if DEBUG:
            print("Entire string is uppercase, thus lowercasing.")
        text = text.lower()
    elif not text_sans_small_words.isupper() and DEBUG:
        print(f"Entire string not upper case. Not lowercasing: {text}")

    lines = re.split(r"[\r\n]+", text)
    processed = []
    for line in lines:
        all_caps = ALL_CAPS.match(line)
        words = re.split(r"[\t ]", line)
        tc_line = []
        for i, word in enumerate(words):
            if DEBUG:
                print(f"Word: {word}")
            if all_caps:
                if UC_INITIALS.match(word):
                    if DEBUG:
                        print(f"  UC_INITIALS match for: {word}")
                    tc_line.append(word)
                    continue
                else:
                    if DEBUG:
                        print(f"  Not initials. Lowercasing: {word}")
                    word = word.lower()

            if APOS_SECOND.match(word):
                # O'Reiley, L'Oreal, D'Angelo
                if DEBUG:
                    print(f"  APOS_SECOND matched. Fixing it: {word}")
                word = word[0:3].upper() + word[3:]
                tc_line.append(word)
                continue

            if INLINE_PERIOD.search(word):
                if DEBUG:
                    print(
                        "  INLINE_PERIOD matched. Uppercasing if == 1 char: "
                        + word
                    )
                parts = word.split(".")
                new_parts = []
                for part in parts:
                    if len(part) == 1:
                        # It's an initial like U.S.
                        new_parts.append(part.upper())
                    else:
                        # It's something like '.com'
                        new_parts.append(part)
                word = ".".join(new_parts)
                tc_line.append(word)
                continue

            if INLINE_SLASH.search(word):
                # This repeats INLINE_PERIOD. Could be more elegant.
                if DEBUG:
                    print(
                        "  INLINE_SLASH matched. Uppercasing if == 1 char: "
                        + word
                    )
                parts = word.split("/")
                new_parts = []
                for part in parts:
                    if len(part) == 1:
                        # It's an initial like A/M
                        new_parts.append(part.upper())
                    else:
                        # It's something like 'True/False'
                        new_parts.append(part)
                word = "/".join(new_parts)
                tc_line.append(word)
                continue

            amp_match = INLINE_AMPERSAND.match(word)
            if amp_match:
                if DEBUG:
                    print(f"  INLINE_AMPERSAND matched. Uppercasing: {word}")
                tc_line.append(
                    f"{amp_match.group(1).upper()}{amp_match.group(2)}"
                )
                continue

            if UC_ELSEWHERE.match(word):
                if DEBUG:
                    print(f"  UC_ELSEWHERE matched. Leaving unchanged: {word}")
                tc_line.append(word)
                continue

            if SMALL_WORDS.match(word):
                if DEBUG:
                    print(f"  SMALL_WORDS matched. Lowercasing: {word}")
                tc_line.append(word.lower())
                continue

            if BIG_WORDS.match(word):
                if DEBUG:
                    print(f"  BIG_WORDS matched. Uppercasing: {word}")
                tc_line.append(word.upper())
                continue

            match = MAC_MC.match(word)
            if match and (word not in ["mack", "machine"]):
                if DEBUG:
                    print(f"  MAC_MAC matched. Capitalizing: {word}")
                tc_line.append(
                    "%s%s"
                    % (
                        match.group(1).capitalize(),
                        match.group(2).capitalize(),
                    )
                )
                continue

            hyphenated = []
            for item in word.split("-"):
                hyphenated.append(_uppercase_word(item))
            tc_line.append("-".join(hyphenated))

        result = " ".join(tc_line)

        result = SMALL_FIRST.sub(
            lambda m: f"{m.group(1)}{m.group(2).capitalize()}",
            result,
        )

        result = SMALL_LAST.sub(lambda m: m.group(0).capitalize(), result)
        result = SUBPHRASE.sub(
            lambda m: f"{m.group(1)}{m.group(2).capitalize()}",
            result,
        )

        processed.append(result)
        text = "\n".join(processed)

    # replace V. with v.
    text = re.sub(re.compile(r"\WV\.\W"), " v. ", text)

    return text


def _uppercase_word(word):
    """
    Helper function for uppercasing a word if it doesn't begin with Unicode characters.

    This is needed due to differences between Python 2 and 3.
    :param word: unicode string to uppercase
    """
    # if six.PY2:
    #    if word[0] in (u'\u2019',):
    #        return CAPFIRST.sub(lambda m: m.group(0).upper(), word[1:])

    return CAPFIRST.sub(lambda m: m.group(0).upper(), word)


def fix_camel_case(s):
    """Sometimes courts provide nasty camel-cased content instead of real
    words. This code attempts to fix that."""
    if " " in s:
        s_out = s
    else:
        s_out = s[0]
        for i in range(1, len(s)):
            # Iterate over the letters, starting with the second one.
            if s[i - 1].isupper() and s[i].isupper():
                # A pattern like 'PAPublic' --> PA Public
                try:
                    next_letter_is_lower = (
                        s[i + 1].islower() and s[i + 1] != "v"
                    )
                except IndexError:
                    # End of string. Break.
                    s_out += s[i]
                    continue
                if next_letter_is_lower:
                    s_out += f" {s[i]}"
                else:
                    s_out += s[i]
            elif s[i - 1].islower() and s[i].isupper():
                # A pattern like 'McLaughlin'
                try:
                    mac_or_mc = (s[i - 2] == "a" and s[i - 3] == "M") or s[
                        i - 2
                    ] == "M"
                except KeyError:
                    s_out += f" {s[i]}"
                if s[i - 1] == "c" and mac_or_mc:
                    s_out += s[i]
                else:
                    s_out += f" {s[i]}"
            elif s[i].isupper() and s[i - 1].islower():
                # Uppercase letter preceded by a lowercase one. Add a space then the letter.
                s_out += f" {s[i]}"
            else:
                s_out += s[i]
        # Fix v.'s that don't have spaces. These come in two forms:
        # 1. So and sov.so and so
        s_out = " ".join(re.sub(r"v\.", " v. ", s_out).split())
        # 2. So and sov so and so --> So and so v. so and so
        if "v." not in s_out:
            s_out = " ".join(re.sub("([^ ])v ", r"\1 v. ", s_out).split())

        # Fix strings like Governorof, Stateof, Secretaryofthe
        if "of" in s_out:
            s_out = re.sub(r"([a-zA-Z])of( |(?:the))", r"\1 of\2", s_out)
        if "the" in s_out:
            s_out = re.sub(r"([a-zA-Z])the ", r"\1 the ", s_out)

        # Fix strings like Santomenno,etal v. John Hancock Life Ins
        if "etal" in s_out:
            s_out = re.sub(r",?etal( |\Z)", " ", s_out).strip()

        # Fix strings like Inre SCH Corp v. CFI Class
        if "Inre" in s_out:
            s_out = re.sub(r"^Inre ", "In Re ", s_out)

    return s_out


# For use in harmonize function
# More details: http://www.law.cornell.edu/citation/4-300.htm
US = r"USA|U\.S\.A\.|U\.S\.?|U\. S\.?|(The )?United States of America|The United States"
UNITED_STATES = re.compile(r"^(%s)(,|\.)?$" % US, re.I)
ET_AL = re.compile(r",?\set\.?\sal\.?", re.I)
BW = (
    "appell(ee|ant)s?|claimants?|complainants?|defendants?|defendants?(--?|/)appell(ee|ant)s?"
    + r"|devisee|executor|executrix|pet(\.|itioner)s?|petitioners?(--?|/)appell(ee|ant)s?"
    + "|petitioners?(--?|/)defendants?|plaintiffs?|plaintiffs?(--?|/)appell(ee|ant)s?|respond(e|a)nts?"
    + "|respond(e|a)nts?(--?|/)appell(ee|ant)s?|cross(--?|/)respondents?|crosss?(--?|/)petitioners?"
    + "|cross(--?|/)appell(ees|ant)s?|deceased"
)
BAD_WORDS = re.compile(r"^(%s)(,|\.)?$" % BW, re.I)


def harmonize(text):
    """Fixes case names so they are cleaner.

    Using a bunch of regex's, this function cleans up common data problems in
    case names. The following are currently fixed:
     - various forms of United States --> United States
     - The State --> State
     - vs. --> v.
     - et al --> Removed.
     - plaintiff, appellee, defendant and the like --> Removed.
     - No. and Nos. removed from beginning

    Lots of tests are in tests.py.
    """
    if not isinstance(text, str):
        text = str(text)

    # replace vs. with v.
    text = re.sub(re.compile(r"\svs\.?\s", re.IGNORECASE), " v. ", text)

    # replace lower case v without a period with v.
    text = re.sub(r"\sv\.?\s", " v. ", text)

    # Remove the BAD_WORDS.
    text = text.split()
    cleaned_text = []
    for word in text:
        word = re.sub(BAD_WORDS, "", word)
        cleaned_text.append(word)
    text = " ".join(cleaned_text)

    # split on all ' v. ' and then deal with United States variations.
    harmonized_parts = []
    parties = text.split(" v. ")
    for party in parties:
        party = party.strip()
        if UNITED_STATES.match(party):
            harmonized_parts.append("United States")
        elif party.lower() == "the state":
            harmonized_parts.append("State")
        else:
            # needed here, because we can't put "US" as a case-insensitive
            # word into the UNITED_STATES regex.
            party = re.sub(re.compile(r"^US$"), "United States", party)
            # no match
            harmonized_parts.append(party)

    result = " v. ".join(harmonized_parts)

    # Remove the ET_AL words.
    result = re.sub(ET_AL, "", result)

    # Fix the No. and Nos.
    if result.startswith("No.") or result.startswith("Nos."):
        result = re.sub(r"^Nos?\.\s+", "", result)

    return clean_string(result)


def clean_string(s):
    """Clean up strings.

    Accomplishes the following:
     - replaces HTML encoded characters with ASCII versions.
     - removes -, ' ', #, *, ; and ',' from the end of lines
     - converts to unicode.
     - removes weird white space and replaces with spaces.
    """
    # if not already unicode, make it unicode, dropping invalid characters
    # if not isinstance(s, unicode):
    s = force_unicode(s, errors="ignore")

    # Get rid of HTML encoded chars
    s = (
        s.replace("&rsquo;", "'")
        .replace("&rdquo;", '"')
        .replace("&ldquo;", '"')
        .replace("&nbsp;", " ")
        .replace("&amp;", "&")
        .replace("%20", " ")
        .replace("&#160;", " ")
    )

    # smart quotes
    s = (
        s.replace("’", "'")
        .replace("‘", "'")
        .replace("“", '"')
        .replace("”", '"')
    )

    # Get rid of weird punctuation
    s = s.replace("*", "").replace("#", "").replace(";", "")

    # Strip bad stuff from the end of lines. Python's strip fails here because
    # we don't know the order of the various punctuation items to be stripped.
    # We split on the v., and handle fixes at either end of plaintiff or
    # appellant.
    bad_punctuation = r"(-|–|_|/|;|,|\s)*"
    bad_endings = re.compile(r"%s$" % bad_punctuation)
    bad_beginnings = re.compile(r"^%s" % bad_punctuation)

    s = s.split(" v. ")
    cleaned_string = []
    for frag in s:
        frag = re.sub(bad_endings, "", frag)
        frag = re.sub(bad_beginnings, "", frag)
        cleaned_string.append(frag)
    s = " v. ".join(cleaned_string)

    # get rid of '\t\n\x0b\x0c\r ', and replace them with a single space.
    s = " ".join(s.split())

    # Convert non-breaking spaces to regular spaces
    s = s.replace("\u00A0", " ")

    return s


def force_unicode(s, encoding="utf-8", strings_only=False, errors="strict"):
    # Borrows heavily from django.utils.encoding.force_unicode.
    # This should be applied to *input* not *output*!
    # Handle the common case first, saves 30-40% in performance when s
    # is an instance of unicode. This function gets called often in that
    # setting.
    if isinstance(s, str):
        return s
    try:
        if not isinstance(s, str):
            if hasattr(s, "__unicode__"):
                s = str(s)
            else:
                try:
                    s = str(str(s), encoding, errors)
                except UnicodeEncodeError:
                    if not isinstance(s, Exception):
                        raise
                    # If we get to here, the caller has passed in an Exception
                    # subclass populated with non-ASCII data without special
                    # handling to display as a string. We need to handle this
                    # without raising a further exception. We do an
                    # approximation to what the Exception's standard str()
                    # output should be.
                    s = " ".join(
                        [
                            force_unicode(arg, encoding, strings_only, errors)
                            for arg in s
                        ]
                    )
        elif not isinstance(s, str):
            # Note: We use .decode() here, instead of unicode(s, encoding,
            # errors), so that if s is a SafeString, it ends up being a
            # SafeUnicode at the end.
            s = s.decode(encoding, errors)
    except UnicodeDecodeError as e:
        if not isinstance(s, Exception):
            raise
        else:
            # If we get to here, the caller has passed in an Exception
            # subclass populated with non-ASCII bytestring data without a
            # working unicode method. Try to handle this without raising a
            # further exception by individually forcing the exception args
            # to unicode.
            s = " ".join(
                [
                    force_unicode(arg, encoding, strings_only, errors)
                    for arg in s
                ]
            )
    return s


def trunc(s, length, ellipsis=None):
    """Truncates a string at a good length.

    Finds the rightmost space in a string, and truncates there. Lacking such
    a space, truncates at length.

    If an ellipsis is provided, the right most space is used that allows the
    addition of the ellipsis without being longer than length.
    """
    if ellipsis:
        ellipsis_length = len(ellipsis)
    else:
        ellipsis_length = 0

    if len(s) <= length:
        # Do not ellipsize if the item is not truncated.
        return s
    else:
        # find the rightmost space using a zero-indexed (+1) length minus the
        # length of the ellipsis.
        rightmost_space_index = length - ellipsis_length + 1
        end = s.rfind(" ", 0, rightmost_space_index)
        if end == -1:
            # no spaces found, just use max position
            end = length - ellipsis_length
        s = s[0:end]
        if ellipsis:
            s = f"{s}{ellipsis}"
        return s


def convert_date_string(date_string, fuzzy=False, datetime=False):
    """Sanitize date string and convert into standard date object

    :param date_string: A string to convert to a datetime object.
    :param fuzzy: whether fuzzy string matching should be used, as defined by
    dateutil.
    :param datetime: If True, return a datetime object. If false, cast to a
    date.
    :return: datetime or date object, depending on the datetime parameter.
    """
    date_string = date_string.replace("(", "")
    date_string = date_string.replace(")", "")

    # python3 and lxml make for odd strings that we need to clean
    date_string = clean_if_py3(date_string)

    date_string = date_string.strip()
    if date_string == "N/A":
        return None

    dt = parser.parse(date_string, fuzzy=fuzzy)
    if datetime:
        return dt
    else:
        return dt.date()


def split_date_range_string(date_range_string):
    """This function requires a string in 'January - March 2016' format"""
    date_range_string = normalize_dashes(date_range_string)
    parts = date_range_string.split()
    if "-" not in date_range_string or len(parts) != 4:
        raise Exception(f'Unrecognized date format: "{date_range_string}"')
    month1, month2, year = parts[0], parts[2], parts[3]
    months = {v: k for k, v in enumerate(calendar.month_name)}
    last_day = calendar.monthrange(int(year), months[month2])[1]
    start_date = convert_date_string(f"{month1} 1, {year}")
    end_date = convert_date_string("%s %d, %s" % (month2, last_day, year))
    delta = end_date - start_date
    dates_in_range = [start_date + timedelta(d) for d in range(delta.days + 1)]
    return dates_in_range[int(len(dates_in_range) / 2)]


def normalize_dashes(raw_string):
    """Replace various dash formats with normal dash"""
    dashes = [
        # copied from http://www.w3schools.com/charsets/ref_utf_punctuation.asp
        "–",  # en dash
        "—",  # em dash
        "‐",  # hyphen
        "‑",  # non-breaking hyphen
        "‒",  # figure dash
        "―",  # horizontal bar
    ]
    normal = "-"
    for dash in dashes:
        raw_string = raw_string.replace(dash, normal)
    return raw_string


class CaseNameTweaker:
    def __init__(self):
        self._bad_words = None
        super().__init__()

    @property
    def bad_words(self):
        """A list of words that shouldn't be in small case names according to
        Blue Book rules.

        Includes:

         - Acronyms
         - Common names
         - Attorneys General
         - Large cities
         - Counties
         - States
         - Punctuation and capitalization variations on the above

        Created lazily and then cached the first time the property is called.
        """
        if self._bad_words is not None:
            return self._bad_words

        acros = [
            "a.g.p.",
            "c.d.c.",
            "c.i.a.",
            "d.o.c.",
            "e.e.o.c.",
            "e.p.a.",
            "f.b.i.",
            "f.c.c.",
            "f.d.i.c.",
            "f.s.b.",
            "f.t.c.",
            "i.c.c.",
            "i.n.s.",
            "i.r.s.",
            "n.a.a.c.p.",
            "n.l.r.b.",
            "p.l.c.",
            "s.e.c.",
            "s.p.a.",
            "s.r.l.",
            "u.s.",
            "u.s.a.",
            "u.s.e.e.o.c.",
            "u.s.e.p.a.",
        ]
        acros_sans_dots = [acro.replace(".", "") for acro in acros]
        # corp_acros = ['L.L.C.', 'L.L.L.P.', 'L.L.P.', 'L.P.', 'P.A.', 'P.C.',
        #              'P.L.L.C.',c ]
        # corp_acros_sans_dots = [acro.replace('.', '') for acro in corp_acros]
        common_names = [
            "state",
            "people",
            "smith",
            "johnson",
            "commissioner",
        ]

        ags = [
            "Akerman",
            "Ashcroft",
            "Barr",
            "Bates",
            "Bell",
            "Berrien",
            "Biddle",
            "Black",
            "Bonaparte",
            "Bork",
            "Bradford",
            "Breckinridge",
            "Brewster",
            "Brownell",
            "Butler",
            "Civiletti",
            "Clark",
            "Clement",
            "Clifford",
            "Crittenden",
            "Cummings",
            "Cushing",
            "Daugherty",
            "Devens",
            "Evarts",
            "Filip",
            "Garland",
            "Gerson",
            "Gilpin",
            "Gonzales",
            "Gregory",
            "Griggs",
            "Grundy",
            "Harmon",
            "Hoar",
            "Holder",
            "Jackson",
            "Johnson",
            "Katzenbach",
            "Keisler",
            "Kennedy",
            "Kleindienst",
            "Knox",
            "Lee",
            "Legaré",
            "Levi",
            "Lincoln",
            "Lynch",
            "MacVeagh",
            "Mason",
            "McGranery",
            "McGrath",
            "McKenna",
            "McReynolds",
            "Meese",
            "Miller",
            "Mitchell",
            "Moody",
            "Mukasey",
            "Murphy",
            "Nelson",
            "Olney",
            "Palmer",
            "Pierrepont",
            "Pinkney",
            "Randolph",
            "Reno",
            "Richardson",
            "Rodney",
            "Rogers",
            "Rush",
            "Sargent",
            "Saxbe",
            "Sessions",
            "Smith",
            "Speed",
            "Stanbery",
            "Stanton",
            "Stone",
            "Taft",
            "Taney",
            "Thornburgh",
            "Toucey",
            "Whitacker",
            "Wickersham",
            "Williams",
            "Wirt",
        ]
        # self.corp_acros = corp_acros + corp_acros_sans_dots
        bad_words = (
            acros
            + acros_sans_dots
            + common_names
            + ags
            + self.make_geographies_list()
        )

        # Add variations with punctuation
        punctuation_bad_words = []
        for word in bad_words:
            for punctuation in string.punctuation:
                if not word.endswith(punctuation):
                    punctuation_bad_words.append(f"{word}{punctuation}")
        bad_words = bad_words + punctuation_bad_words

        bad_words = [s.lower() for s in bad_words]

        self._bad_words = bad_words
        return bad_words

    @staticmethod
    def make_geographies_list():
        """Make a flat list of cities, counties and states that we can exclude
        from short names.
        """
        geonames = geonamescache.GeonamesCache()

        # Make a list of cities with big populations.
        cities = [
            v["name"]
            for v in geonames.get_cities().values()
            if (v["countrycode"] == "US" and v["population"] > 150000)
        ]
        counties = [v["name"] for v in geonames.get_us_counties()]
        states = [v["name"] for v in geonames.get_us_states().values()]
        return cities + counties + states

    def make_case_name_short(self, s):
        """Creates short case names where obvious ones can easily be made."""
        parts = [part.strip().split() for part in s.split(" v. ")]
        if len(parts) == 1:
            # No v. Likely an "In re" or "Matter of" case.
            if len(parts[0]) <= 3:
                # Good length for a shortened case name.
                return s
            else:
                # Too long; too weird. Punt.
                return ""

        elif len(parts) == 2:
            # X v. Y --> [['X'], ['Y']]
            # X Y Z v. A B --> [['X', 'Y', 'Z'], ['A', 'B']]
            if len(parts[0]) == 1:
                if parts[0][0].lower() not in self.bad_words:
                    # Simple case: Langley v. Google
                    return parts[0][0]
                elif len(parts[1]) == 1:
                    # Plaintiff was a bad_word. Try the defendant.
                    # Dallas v. Lissner
                    if parts[1][0].lower() not in self.bad_words:
                        return parts[1][0]
            elif len(parts[0]) > 1:
                # Plaintiff part is longer than a single word, focus on the
                # defendant.
                if len(parts[1]) == 1:
                    # If the defendant is a single word.
                    if parts[1][0].lower() not in self.bad_words:
                        # That's not a bad word.
                        return parts[1][0]

        # More than 1 instance of v. or otherwise no matches --> Give up.
        return ""


def clean_if_py3(s):
    """
    Cleans up text if using Python 3

    :param s: string to clean
    :return: string un-modified on Python 2.x, string with raw literals replaced
             with their string forms on Python 3.x
    """
    replacements = [(r"\n", "\n"), (r"\t", "\t"), (r"\r", "\r")]

    for raw, replacement in replacements:
        s = s.replace(raw, replacement)

    return s



"""
Scraper for Court of Appeals of Arizona, Division 2
CourtID: arizctapp_div_2
Court Short Name: arizctapp_div_2
Author: Andrei Chelaru
Reviewer: mlr
History:
    2014-07-23: Created by Andrei Chelaru
    2021-12-10: URL changed to recent opinions page, satsuki-chan
"""

from juriscraper.lib.string_utils import clean_if_py3, titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        # Feeling down and tired of of your regular life? Check out this website.
        self.url = "https://www.appeals2.az.gov/ODSPlus/recentOpinionsHTML.cfm"
        self.cases = []

    def _process_html(self) -> None:
        path = "//table//a[contains(@href, '.pdf')]"
        for item in self.html.xpath(path):
            docket = item.xpath("./text()")[0]
            url = item.xpath("./@href")[0]
            name = item.xpath("./following::td[1]/*/text()")[0]
            # Expected: "Opinion Filed: 10/13/2021"
            date_string = item.xpath("./following::td[2]/text()")[0].strip()
            date = clean_if_py3(date_string).rsplit(":", 1)[1].strip()
            summary_list = item.xpath("./following::tr[1]//text()")
            summary = "".join(summary_list).strip()
            self.cases.append(
                {
                    "name": titlecase(name),
                    "date": date,
                    "docket": docket,
                    "url": url,
                    "summary": summary,
                    "status": "Published",
                }
            )



"""Scraper for the Maryland Attorney General
CourtID: mdag
Court Short Name: Maryland Attorney General
"""

import json
import re
from datetime import date

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = (
            "https://www.marylandattorneygeneral.gov/_layouts/15/inplview.aspx"
        )
        self.year = date.today().year

        self.parameters = {
            "List": "{1BA692B1-E50C-4754-AADD-E6753F46B403}",
            "View": "{E1A60D10-12C0-4029-8BE0-BA5F9AC93BF8}",
            "ViewCount": "50",  # results to return
            "IsXslView": "TRUE",
            "IsCSR": "TRUE",
            "GroupString": f";#{self.year};#",
            "IsGroupRender": "TRUE",
        }
        self.status = "Published"

    def _download(self, request_dict={}):
        if self.test_mode_enabled():
            with open(self.url) as file:
                self.json = json.load(file)
        else:
            self.json = (
                self.request["session"]
                .post(self.url, params=self.parameters)
                .json()
            )

    def _process_html(self):
        for row in self.json["Row"]:
            url_path = row["FileRef.urlencodeasurl"]

            # created_x0020_date looks like this 0;#2021-11-09 14:29:09
            datetime_string = row["Created_x0020_Date"]
            date = re.findall(r"\d{4}-\d{2}-\d{2}", datetime_string)[0]
            self.cases.append(
                {
                    "docket": row["Title"],
                    "summary": row["Summary"],
                    "name": row["FileLeafRef.Name"],
                    "url": f"https://www.marylandattorneygeneral.gov{url_path}",
                    "date": date,
                }
            )



import argparse
import os
import pprint
import re
import sys
from html import unescape

import feedparser
from requests import Session

from ..lib.log_tools import make_default_logger
from ..lib.string_utils import clean_string, harmonize
from ..lib.utils import clean_court_object
from .docket_report import DocketReport
from .utils import (
    get_pacer_case_id_from_nonce_url,
    get_pacer_doc_id_from_doc1_url,
    get_pacer_seq_no_from_doc1_url,
    parse_datetime_for_us_timezone,
    set_pacer_doc_id_as_appellate_document_number,
)

logger = make_default_logger()

"""
As of 2018-04-25, the jurisdictions below do not have functional RSS feeds. I
reached out to all of these jurisdictions and heard back the following:

 - miwb: Sent email. Left a VM.
 - nceb: "The Judges in our District have given direction that they do not want
   the RSS feed turned on in our court.  This has been discussed several times
   and their decision remains the same."
    - Update via email: "At this time, the court declined to participate.
      However, the court anticipates transitioning to the NextGen version of
      CM/ECF in 2019 and will likely participate at that time.
 - alnd: Sent email.
 - caed: Sent email. 916-930-4000, left message with IT director, Richard.
 - flnd: Sent email.
 - gand: Left a message at the "Systems" department
 - gasd: Sent email.
 - hid: Sent email. 808-541-1304, Lian Abernathy, Chief Deputy
 - ilsd: Sent email.
 - kyed: Need to call Sharon Drolijk (Operations Manager) at (859) 233-2503.
 - mdd: Spoke with "Attorney Adviser". He asked for a letter, which I sent. He
   will take it up with the court, though they didn't have any policy that he
   was aware of. It's going through their "court committee system for
   consideration." Seems it wasn't done for performance reasons years ago. They
   will take it up again.
 - msnd: Sent email.
 - mtd: Julie Collins, Deputy Clerk in charge of Billings division, called back.
   She's going to look into it.
 - ndd: Sent email. 701-530-2313, Clerk of Clerk, Rob. On the to do list.
   Hopefully by end of the year (2018). Nobody ever asked before.
 - nvd: Sent email. Talked to Robert in "Quality Control". He shook me a bit
   with how difficult he was. He insists I need to send a letter to the chief
   judge, but that they're not considering this currently.
    - A different tack: Trying the operations manager, Lia at 775-686-5840.
      She's friendly. Requested a letter for chief judge. Sent.
 - nywd: Sent email.
 - oked: Sent email.
 - oknd: Sent email.
 - pamd: Sent email.
 - scd: "We have forwarded your request on to management staff at the District
   Court for further consideration."
 - tnwd: Sent email.
 - txnd: Left me a long voicemail. They're working on it and have it in
   committee. Expectation is that they may require an en banc meeting of the
   judges to make a decision, but that the decision should come soon and be
   positive.
 - vaed: "We are currently looking into this and will possibly have this
   feature in the near future."

"""


def append_or_merge_entry(docket_list, new_docket):
    """Append new entry to our output or merge it if it's a multi-event entry.

    CMECF entries can contain multiple events, e.g. anyone filing a Motion can
    select one or more of the event types; so too for judicial orders. When
    this is rendered in the RSS feed, it is rendered as two (or more) entries
    for the same case document, with identical times, titles, links, and guids.

    If the new entry is part of a multi-event entry, merge it with the prior
    one. If it's not, append it to the list.

    :param docket_list: A list of docket-like dictionaries, each containing
    docket entries
    :param new_docket: A new docket-like dictionary that can be appended or
    merged into the docket_list.
    :return None
    """
    for docket in docket_list:
        entry = docket["docket_entries"][0]
        new_entry = new_docket["docket_entries"][0]
        same_dn = docket["docket_number"] == new_docket["docket_number"]
        same_cn = docket["pacer_case_id"] == new_docket["pacer_case_id"]
        same_date = entry["date_filed"] == new_entry["date_filed"]
        same_doc_id = entry["pacer_doc_id"] == new_entry["pacer_doc_id"]
        if all([same_dn, same_cn, same_date, same_doc_id]):
            # if docket number, pacer_case_id, date filing, and pacer_doc_id
            # are same, order short descriptions alphabetically and merge.

            short_descriptions = [
                desc.strip()
                for desc in entry["short_description"].split("AND")
            ]
            short_descriptions.append(new_entry["short_description"])
            short_description = " AND ".join(sorted(short_descriptions))
            entry["short_description"] = short_description
            break
    else:
        # Loop exited without hitting a break; item is distinct; append.
        docket_list.append(new_docket)


class PacerRssFeed(DocketReport):
    # The entries are HTML entity-coded, and these matches are run AFTER
    # decoding. A typical entry is of the form:
    #   [short_description] (<a href="doc1_url">document_number</a>)
    # Or a literal example (line breaks added):
    #   [Scheduling Order] (<a href="https://ecf.mad.uscourts.gov/doc1/
    #   09518690740?caseid=186403&de_seq_num=98">39</a>)
    # We use three simple matches rather than a complex one with three groups.
    document_number_regex = re.compile(r"[\"|\'] ?>(\d+)</a>")
    doc1_url_regex = re.compile(r"href=[\"|\'](.*/doc1/.*)[\"|\']")
    docs1_url_regex = re.compile(r"href=[\"|\'](.*/docs1/.*)[\"|\']")
    short_desc_regex = re.compile(r"\[(.*?)\]")  # Matches 'foo': [ foo ] (

    PATH = "cgi-bin/rss_outside.pl"

    CACHE_ATTRS = ["data"]

    def __init__(self, court_id):
        super().__init__(court_id)
        self._clear_caches()
        self._data = None
        self.session = Session()
        self.is_valid = True
        self.is_appellate = False
        if self.court_id[-1].isdigit() or self.court_id in [
            "cadc",
            "cafc",
            "cavc",
        ]:
            self.is_appellate = True
        if self.court_id.endswith("b"):
            self.is_bankruptcy = True
        else:
            self.is_bankruptcy = False

    def __del__(self):
        if self.session:
            self.session.close()

    @property
    def url(self):
        if self.court_id == "nyed":
            # EDNY has some special RSS software. In Sept. 2017,
            # rss_outside.pl redirected to readyDockets.pl:
            #   https://twitter.com/johnhawkinson/status/906005915622629376
            # But that had stopped happening by 2020.
            # Special case it for now, but EDNY/AOUSC follow is required:
            return "https://ecf.nyed.uscourts.gov/cgi-bin/readyDockets.pl"
        elif self.court_id == "ared":
            # In response to requests from their bar association, the Eastern
            # District of Arkansas has a bunch of RSS feeds that you can see
            # from their homepage. Their default one is limited to orders and
            # opinions, but this one has "All Docket Entries".
            return "https://ecf.ared.uscourts.gov/cgi-bin/rss_outside4.pl"
        elif self.is_appellate:
            return f"https://ecf.{self.court_id}.uscourts.gov/n/beam/servlet/TransportRoom?servlet=RSSGenerator"
        else:
            return f"https://ecf.{self.court_id}.uscourts.gov/{self.PATH}"

    def query(self):
        """Query the RSS feed for a given court ID

        Note that we use requests here, and so we forgo some of the
        useful features that feedparser offers around the Etags and
        Last-Modified headers. This is fine for now because no PACER
        site seems to support these headers, but eventually we'll
        probably want to do better here. The reason we *don't* use
        feedparser already is because it presents a different set of
        APIs and Exceptions that we don't want to monkey with, at
        least, not yet, and especially not yet if PACER itself doesn't
        seem to care.

        For a good summary of this issue, see:
        https://github.com/freelawproject/juriscraper/issues/195#issuecomment-385848344
        """
        logger.info(f"Querying the RSS feed for {self.court_id}")
        # The timeout here is a bit tricky. Too long, and national PACER
        # outages cause us grief. Too short and slow courts don't get done.
        # Previously, this value has been (60, 300), then 5. Hopefully the
        # below is a reasonable middle ground.
        timeout = (5, 20)
        self.response = self.session.get(self.url, timeout=timeout)

    def parse(self):
        self._clear_caches()
        self.response.raise_for_status()
        self._parse_text(self.response.text)

    def _parse_text(self, text):
        """Parse the feed and set self.feed

        :param text: The text of the RSS feed.
        :return None
        """
        self.feed = feedparser.parse(text)

    @property
    def data(self):
        """Override this to create a list of docket-like objects instead of the
        usual dict that is usually provided by the docket report.
        """
        if self._data is not None:
            return self._data

        docket_list = []
        for entry in self.feed.entries:
            try:
                new_docket = self.metadata(entry)
                new_docket["parties"] = None
                new_docket["docket_entries"] = self.docket_entries(entry)
            except AttributeError:
                # Happens when RSS items lack necessary attributes like a URL
                # or published date.
                pass
            else:
                if (
                    new_docket["docket_entries"]
                    and new_docket["docket_number"]
                ):
                    append_or_merge_entry(docket_list, new_docket)

        self._data = docket_list
        return docket_list

    def metadata(self, entry):
        if self.is_valid is False:
            return {}

        data = {
            "court_id": self.court_id,
            "pacer_case_id": get_pacer_case_id_from_nonce_url(entry.link)
            if not self.is_appellate
            else None,
            "docket_number": self._get_docket_number(entry.title),
            "case_name": self._get_case_name(entry.title),
            # Filing date is not available. Also the case for free opinions.
            "date_filed": None,
            "date_terminated": None,
            "date_converted": None,
            "date_discharged": None,
            "assigned_to_str": "",
            "referred_to_str": "",
            "cause": "",
            "nature_of_suit": "",
            "jury_demand": "",
            "demand": "",
            "jurisdiction": "",
            # bankruptcy data
            "trustee_str": "",
            "type": "",
            "office": "",
            "chapter": "",
        }
        data.update(self._parse_get_trustee_type_office_chapter(entry.summary))
        data = clean_court_object(data)
        return data

    @property
    def parties(self):
        raise NotImplementedError("No parties for RSS feeds.")

    def docket_entries(self, entry):
        """Parse the RSS item to get back a docket entry-like object.
        Although there is only one, return it as a list.
        """
        de = {
            "date_filed": parse_datetime_for_us_timezone(entry.published),
            "description": "",
            "document_number": self._get_value(
                self.document_number_regex, entry.summary
            )
            or None,
            "short_description": unescape(
                self._get_value(self.short_desc_regex, entry.summary)
            ),
        }

        # District and bankruptcy doc URL
        doc1_url = self._get_value(self.doc1_url_regex, entry.summary)
        # Appellate doc URL
        docs1_url = self._get_value(self.docs1_url_regex, entry.summary)
        if doc1_url:
            de["pacer_doc_id"] = get_pacer_doc_id_from_doc1_url(doc1_url)
            de["pacer_seq_no"] = get_pacer_seq_no_from_doc1_url(doc1_url)
        elif docs1_url:
            de["pacer_doc_id"] = get_pacer_doc_id_from_doc1_url(docs1_url)
            de["pacer_seq_no"] = None
        else:
            # Some courts, in particular, NYED do not provide doc1 links and
            # instead provide show_case_doc links. Some docket entries don't
            # provide links. In either case, we can't provide pacer_doc_id.
            de["pacer_doc_id"] = ""
            de["pacer_seq_no"] = None

        if self.is_appellate:
            set_pacer_doc_id_as_appellate_document_number(de)

        return [de]

    def _parse_get_trustee_type_office_chapter(self, entry_text):
        if not self.is_bankruptcy:
            return {}
        data_parts_re = re.compile(
            r"""
            ^Type: (?P<type>.*?)
            Office: (?P<office>.*?)
            Chapter: (?P<chapter>.*?)
            (Trustee: (?P<trustee>.*?))? \[""",
            re.VERBOSE,
        )
        m = data_parts_re.search(entry_text)
        return {
            "type": m.group("type"),
            "office": m.group("office"),
            "chapter": m.group("chapter"),
            "trustee_str": m.group("trustee") or "",
        }

    def _get_docket_number(self, title_text):
        if self.is_bankruptcy:
            # Uses both b/c sometimes the bankr. cases have a dist-style docket
            # number.
            regexes = [
                self.docket_number_dist_regex,
                self.docket_number_bankr_regex,
            ]
        elif self.is_appellate:
            regexes = [self.docket_number_appellate_regex]
        else:
            regexes = [self.docket_number_dist_regex]
        for regex in regexes:
            match = regex.search(title_text)
            if match:
                return match.group(1)

    def _get_case_name(self, title_text):
        # 1:18-cv-04423 Chau v. Gorg &amp; Smith et al --> Chau v. Gorg & Smith
        try:
            case_name = title_text.split(" ", 1)[1]
        except IndexError:
            return "Unknown Case Title"
        case_name = unescape(case_name)
        case_name = clean_string(harmonize(case_name))
        return case_name


def _main():
    # For help: python -m juriscraper.pacer.rss_feeds -h
    parser = argparse.ArgumentParser(
        prog="python -m %s.%s"
        % (__package__, os.path.splitext(os.path.basename(sys.argv[0]))[0])
    )
    parser.add_argument(
        "-b",
        "--bankruptcy",
        action="store_true",
        help="Use bankruptcy parser variant.",
    )
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument(
        "court_or_file",
        nargs="?",
        default="-",
        help="""
A PACER court id or a local filename; defaults to stdin (-).
Any 3 or 4 character string is presumed to be a court;
sorry if that was your filename.""",
    )

    args = parser.parse_args()

    arg_len = len(args.court_or_file)
    if 3 <= arg_len <= 4:
        feed = PacerRssFeed(args.court_or_file)
        print(f"Querying RSS feed at: {feed.url}")
        feed.query()
        print(f"Parsing RSS feed for {feed.court_id}")
        feed.parse()
    else:
        if not args.bankruptcy:
            feed = PacerRssFeed("fake_district_court_id")
        else:
            # final 'b' char is interpretted as bankruptcy
            feed = PacerRssFeed("fake_bankruptcy_court_id_b")
        if args.court_or_file == "-":
            print(f"Faking up RSS feed from stdin as {feed.court_id}")
            f = sys.stdin
        else:
            print(
                "Reading RSS feed from %s as %s"
                % (args.court_or_file, feed.court_id)
            )
            f = open(args.court_or_file)
        feed._parse_text(f.read().decode("utf-8"))

    print(f"Got {len(feed.data)} items")
    if args.verbose:
        print("Here they are:\n")
        pprint.pprint(feed.data, indent=2)


if __name__ == "__main__":
    _main()



"""
Scraper for the United States Bankruptcy Appellate Panel for the Ninth Circuit
CourtID: bap9
Court Short Name: 9th Cir. BAP
"""

from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "https://www.ca9.uscourts.gov/bap/"
        self.court_id = self.__module__
        self.parameters = {
            "c_mode": "view",
            "c_page_size": "500",
        }

    def _process_html(self):
        path = "//table[@id='search-results-table']//tr[position()>1]"
        for row in self.html.xpath(path):
            cell_1 = row.xpath("./td[1]")[0]
            url = cell_1.xpath(".//a/@href")
            if not url:
                continue
            type = row.xpath("./td[2]")[0].text_content()
            if "Unpublished" in type:
                status = "Unpublished"
            elif "Published" in type:
                status = "Published"
            else:
                status = "Unknown"
            self.cases.append(
                {
                    "url": url[0],
                    "name": f"In re: {titlecase(cell_1.text_content())}",
                    "status": status,
                    "docket": row.xpath("./td[3]")[0].text_content(),
                    "date": row.xpath("./td[4]")[0].text_content(),
                }
            )



from juriscraper.opinions.united_states.state import cal


class Site(cal.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "https://www.courts.ca.gov/cms/npopinions.htm"
        self.court_id = self.__module__

    def build_url(self):
        return "https://www.courts.ca.gov/cms/npopinions.htm"

    def _get_precedential_statuses(self):
        return ["Unpublished"] * len(self.case_names)



"""Scraper for [Full name of court]
CourtID: [unique abbreviation to be used by software/filesystem]
Court Short Name: [standard abbreviation used in citations]
Author:
Reviewer:
History:
  YYYY-MM-DD: Created by XXX
"""

from lxml import html

from juriscraper.lib.string_utils import convert_date_string, titlecase
from juriscraper.OralArgumentSite import OralArgumentSite


class Site(OralArgumentSite):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "http://court-url.gov/some-path.html"
        self.method = "POST"
        self.uses_selenium = False
        # Complete this variable if you create a backscraper.
        self.back_scrape_iterable = None

    """
      Required fields - InsanityExceptions will be thrown if omitted.

      Remove this comment before submission.
    """

    def _get_download_urls(self):
        """This is an example of a basic piece of meta data accessible with a
        simple XPath query.

        On a good site, most of your methods will follow this pattern.

        Note that relative URLs are made absolute by the AbstractSite
        object's cleanup routines, so doing so here is not needed.
        """
        path = "//path/to/text/text()"
        return list(self.html.xpath(path))

    def _get_case_names(self):
        """This example demonstrates how to extract text from an element that
        may contain other elements.

        For example, this will work well on something like:
           <strong>Nadim v. <em>Jenny</em></strong>

        Resulting in text like:
           Nadim v. Jenny

        Note that titlecase() should be used here in the case that the case
        names are provided in uppercase. Use the titlecase function on
        cases where the name is provided in uppercase only.
        """
        case_names = []
        for e in self.html.xpath("//path/to/an/element/p"):
            s = html.tostring(e, method="text", encoding="unicode")
            case_names.append(titlecase(s))
        return case_names

    def _get_case_dates(self):
        """This is an example of a date field. Note that the format string
        will likely need to be updated to  match the date formats
        on the site you are scraping. The datetime formats can be found
        here: http://docs.python.org/2/library/datetime.html
        """
        path = "//path/to/text/text()"
        return [
            convert_date_string(date_string)
            for date_string in self.html.xpath(path)
        ]

    """
      High priority fields

      Remove this comment and any unused methods before submission
    """

    def _get_docket_numbers(self):
        """
        This is typically of the form ##-####
        """
        return None

    def _get_judges(self):
        """
        For the name of the judge authoring the lead opinion
        """
        return None

    """
      Optional method used for downloading multiple pages of a court site.
    """

    def _download_backwards(self, date_str):
        """This is a simple method that can be used to generate Site objects
        that can be used to paginate through a court's entire website.

        This method is usually called by a backscraper caller (see the
        one in CourtListener/alert/scrapers for details), and typically
        modifies aspects of the Site object's attributes such as Site.url.

        A simple example has been provided below. The idea is that the
        caller runs this method with a different variable on each iteration.
        That variable is often a date that is getting iterated or is simply
        a index (i), that we iterate upon.

        This can also be used to hold notes useful to future backscraper
        development.
        """
        self.url = f"http://example.com/new/url/{date_str}"
        self.html = self._download()



"""Scraper for Dept of Justice Office of Legal Counsel
CourtID: bia
Court Short Name: Dept of Justice OLC
Author: William Palin
Reviewer:
Type:
History:
    2022-01-14: Created by William E. Palin
"""

from juriscraper.lib.html_utils import (
    get_row_column_links,
    get_row_column_text,
)
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://www.justice.gov/olc/opinions?items_per_page=40"
        self.status = "Published"

    def _process_html(self):
        for row in self.html.xpath(
            ".//tr[contains(@class , 'even')] | .//tr[contains(@class , 'odd')]"
        ):
            date = get_row_column_text(row, 1)
            if "Date of Issuance" in date:
                date = date.split("\n")[-1].strip()
            name = get_row_column_text(row, 2)
            url = get_row_column_links(row, 2)
            summary = get_row_column_text(row, 3)
            self.cases.append(
                {
                    "date": date,
                    "name": name,
                    "url": url,
                    "summary": summary,
                    "docket": "",  # Docket numbers don't appear to exist.
                }
            )



from juriscraper.opinions.united_states.federal_appellate import scotus_slip


class Site(scotus_slip.Site):
    # Note that scotus_relating inherits from this class.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.back_scrape_iterable = list(range(5, int(self.yy) + 1))
        self.precedential = "In-chambers"
        self.court = "in-chambers"

    def set_url(self):
        self.url = f"{self.url_base}/{self.court}.aspx"



import re
from datetime import date, datetime
from typing import Dict, Optional, Union

import requests
import tldextract
from dateutil import parser
from dateutil.tz import gettz
from lxml import html

from ..lib.exceptions import ParsingException


def get_court_id_from_doc_id_prefix(prefix):
    prefix_to_cid_map = {
        "012": "cavc",
        "014": "cit",
        "015": "cofc",
        "016": "almb",
        "017": "almd",
        "018": "alnb",
        "019": "alnd",
        "020": "alsb",
        "021": "alsd",
        "022": "akb",
        "023": "akd",
        "024": "arb",
        "025": "azd",
        "026": "areb",
        "027": "ared",
        "028": "arwb",
        "029": "arwd",
        "031": "cacd",
        "032": "caeb",
        "033": "caed",
        "034": "canb",
        "035": "cand",
        "036": "casb",
        "037": "casd",
        "038": "cob",
        "039": "cod",
        "040": "ctb",
        "041": "ctd",
        "042": "deb",
        "043": "ded",
        "044": "dcb",
        "045": "dcd",
        "046": "flmb",
        "047": "flmd",
        "048": "flnb",
        "049": "flnd",
        "050": "flsb",
        "051": "flsd",
        "052": "gamb",
        "053": "gamd",
        "054": "ganb",
        "055": "gand",
        "056": "gasb",
        "057": "gasd",
        "058": "gub",
        "059": "gud",
        "060": "hib",
        "061": "hid",
        "062": "idb",
        "063": "idd",
        "064": "ilcb",
        "065": "ilcd",
        "066": "ilnb",
        "067": "ilnd",
        "068": "ilsb",
        "069": "ilsd",
        "070": "innb",
        "071": "innd",
        "072": "insb",
        "073": "insd",
        "074": "ianb",
        "075": "iand",
        "076": "iasb",
        "077": "iasd",
        "078": "ksb",
        "079": "ksd",
        "080": "kyeb",
        "081": "kyed",
        "082": "kywb",
        "083": "kywd",
        "084": "laeb",
        "085": "laed",
        "086": "lamb",
        "087": "lamd",
        "088": "lawb",
        "089": "lawd",
        "090": "meb",
        "091": "med",
        "092": "mdb",
        "093": "mdd",
        "094": "mab",
        "095": "mad",
        "096": "mieb",
        "097": "mied",
        "098": "miwb",
        "099": "miwd",
        "100": "mnb",
        "101": "mnd",
        "102": "msnb",
        "103": "msnd",
        "104": "mssb",
        "105": "mssd",
        "106": "moeb",
        "107": "moed",
        "108": "mowb",
        "109": "mowd",
        "110": "mtb",
        "111": "mtd",
        "112": "neb",
        "113": "ned",
        "114": "nvb",
        "115": "nvd",
        "116": "nhb",
        "117": "nhd",
        "118": "njb",
        "119": "njd",
        "120": "nmb",
        "121": "nmd",
        "122": "nyeb",
        "123": "nyed",
        "124": "nynb",
        "125": "nynd",
        "126": "nysb",
        "127": "nysd",
        "128": "nywb",
        "129": "nywd",
        "130": "nceb",
        "131": "nced",
        "132": "ncmb",
        "133": "ncmd",
        "134": "ncwb",
        "135": "ncwd",
        "136": "ndb",
        "137": "ndd",
        "138": "nmib",
        "139": "nmid",
        "140": "ohnb",
        "141": "ohnd",
        "142": "ohsb",
        "143": "ohsd",
        "144": "okeb",
        "145": "oked",
        "146": "oknb",
        "147": "oknd",
        "148": "okwb",
        "149": "okwd",
        "150": "orb",
        "151": "ord",
        "152": "paeb",
        "153": "paed",
        "154": "pamb",
        "155": "pamd",
        "156": "pawb",
        "157": "pawd",
        "158": "prb",
        "159": "prd",
        "160": "rib",
        "161": "rid",
        "162": "scb",
        "163": "scd",
        "164": "sdb",
        "165": "sdd",
        "166": "tneb",
        "167": "tned",
        "168": "tnmb",
        "169": "tnmd",
        "170": "tnwb",
        "171": "tnwd",
        "174": "txeb",
        "175": "txed",
        "176": "txnb",
        "177": "txnd",
        "178": "txsb",
        "179": "txsd",
        "180": "txwb",
        "181": "txwd",
        "182": "utb",
        "183": "utd",
        "184": "vtb",
        "185": "vtd",
        "188": "vaeb",
        "189": "vaed",
        "190": "vawb",
        "191": "vawd",
        "192": "vib",
        "193": "vid",
        "194": "waeb",
        "195": "waed",
        "196": "wawb",
        "197": "wawd",
        "198": "wvnb",
        "199": "wvnd",
        "200": "wvsb",
        "201": "wvsd",
        "202": "wieb",
        "203": "wied",
        "204": "wiwb",
        "205": "wiwd",
        "206": "wyb",
        "207": "wyd",
        "850": "jpml",
        "973": "cacb",
    }
    return prefix_to_cid_map[prefix]


def get_doc_id_prefix_from_court_id(court_id):
    cid_to_prefix_map = {
        "akb": "022",
        "akd": "023",
        "almb": "016",
        "almd": "017",
        "alnb": "018",
        "alnd": "019",
        "alsb": "020",
        "alsd": "021",
        "arb": "024",
        "areb": "026",
        "ared": "027",
        "arwb": "028",
        "arwd": "029",
        "azd": "025",
        "cacb": "973",
        "cacd": "031",
        "caeb": "032",
        "caed": "033",
        "canb": "034",
        "cand": "035",
        "casb": "036",
        "casd": "037",
        "cavc": "012",
        "cit": "014",
        "cob": "038",
        "cod": "039",
        "cofc": "015",
        "ctb": "040",
        "ctd": "041",
        "dcb": "044",
        "dcd": "045",
        "deb": "042",
        "ded": "043",
        "flmb": "046",
        "flmd": "047",
        "flnb": "048",
        "flnd": "049",
        "flsb": "050",
        "flsd": "051",
        "gamb": "052",
        "gamd": "053",
        "ganb": "054",
        "gand": "055",
        "gasb": "056",
        "gasd": "057",
        "gub": "058",
        "gud": "059",
        "hib": "060",
        "hid": "061",
        "ianb": "074",
        "iand": "075",
        "iasb": "076",
        "iasd": "077",
        "idb": "062",
        "idd": "063",
        "ilcb": "064",
        "ilcd": "065",
        "ilnb": "066",
        "ilnd": "067",
        "ilsb": "068",
        "ilsd": "069",
        "innb": "070",
        "innd": "071",
        "insb": "072",
        "insd": "073",
        "jpml": "850",
        "ksb": "078",
        "ksd": "079",
        "kyeb": "080",
        "kyed": "081",
        "kywb": "082",
        "kywd": "083",
        "laeb": "084",
        "laed": "085",
        "lamb": "086",
        "lamd": "087",
        "lawb": "088",
        "lawd": "089",
        "mab": "094",
        "mad": "095",
        "mdb": "092",
        "mdd": "093",
        "meb": "090",
        "med": "091",
        "mieb": "096",
        "mied": "097",
        "miwb": "098",
        "miwd": "099",
        "mnb": "100",
        "mnd": "101",
        "moeb": "106",
        "moed": "107",
        "mowb": "108",
        "mowd": "109",
        "msnb": "102",
        "msnd": "103",
        "mssb": "104",
        "mssd": "105",
        "mtb": "110",
        "mtd": "111",
        "nceb": "130",
        "nced": "131",
        "ncmb": "132",
        "ncmd": "133",
        "ncwb": "134",
        "ncwd": "135",
        "ndb": "136",
        "ndd": "137",
        "neb": "112",
        "ned": "113",
        "nhb": "116",
        "nhd": "117",
        "njb": "118",
        "njd": "119",
        "nmb": "120",
        "nmd": "121",
        "nmib": "138",
        "nmid": "139",
        "nvb": "114",
        "nvd": "115",
        "nyeb": "122",
        "nyed": "123",
        "nynb": "124",
        "nynd": "125",
        "nysb": "126",
        "nysd": "127",
        "nywb": "128",
        "nywd": "129",
        "ohnb": "140",
        "ohnd": "141",
        "ohsb": "142",
        "ohsd": "143",
        "okeb": "144",
        "oked": "145",
        "oknb": "146",
        "oknd": "147",
        "okwb": "148",
        "okwd": "149",
        "orb": "150",
        "ord": "151",
        "paeb": "152",
        "paed": "153",
        "pamb": "154",
        "pamd": "155",
        "pawb": "156",
        "pawd": "157",
        "prb": "158",
        "prd": "159",
        "rib": "160",
        "rid": "161",
        "scb": "162",
        "scd": "163",
        "sdb": "164",
        "sdd": "165",
        "tneb": "166",
        "tned": "167",
        "tnmb": "168",
        "tnmd": "169",
        "tnwb": "170",
        "tnwd": "171",
        "txeb": "174",
        "txed": "175",
        "txnb": "176",
        "txnd": "177",
        "txsb": "178",
        "txsd": "179",
        "txwb": "180",
        "txwd": "181",
        "utb": "182",
        "utd": "183",
        "vaeb": "188",
        "vaed": "189",
        "vawb": "190",
        "vawd": "191",
        "vib": "192",
        "vid": "193",
        "vtb": "184",
        "vtd": "185",
        "waeb": "194",
        "waed": "195",
        "wawb": "196",
        "wawd": "197",
        "wieb": "202",
        "wied": "203",
        "wiwb": "204",
        "wiwd": "205",
        "wvnb": "198",
        "wvnd": "199",
        "wvsb": "200",
        "wvsd": "201",
        "wyb": "206",
        "wyd": "207",
    }
    return cid_to_prefix_map[court_id]


def get_pacer_court_info():
    r = requests.get("https://court-version-scraper.fly.dev/courts.json")
    return r.json()


def get_courts_from_json(j):
    courts = []
    for k, v in j.items():
        for court in v["courts"]:
            court["type"] = k
            courts.append(court)
    return courts


def get_court_id_from_url(url):
    """Extract the court ID from the URL."""
    parts = tldextract.extract(url)
    return parts.subdomain.split(".")[1]


def get_pacer_case_id_from_nonce_url(url):
    """Extract the pacer case ID from the URL.

    In: https://ecf.almd.uscourts.gov/cgi-bin/DktRpt.pl?56120
    Out: 56120
    In: https://ecf.azb.uscourts.gov/cgi-bin/iquery.pl?625371913403797-L_9999_1-0-663150
    Out: 663150
    """
    param = url.split("?")[1]
    if "L" in param:
        return param.rsplit("-", 1)[1]
    return param


def get_pacer_seq_no_from_doc1_url(url):
    """Extract the seq_no from the doc1 URL."""
    match = re.search(r"de_seq_num=(\d+)", url)
    if match:
        return match.group(1)
    else:
        return None


def get_pacer_case_id_from_doc1_url(url):
    """Extract the caseid from the doc1 URL."""
    match = re.search(r"caseid=(\d+)", url)
    if match:
        return match.group(1)
    else:
        return None


def get_pacer_magic_num_from_doc1_url(
    url: str,
    appellate: bool = False,
) -> Optional[str]:
    """Extract the magic number from the doc1 URL."""
    if appellate:
        # NDA free look link format is:
        # https://ecf.ca2.uscourts.gov/docs1/00208721516?uid=b775e9908ad79ce2
        match = re.search(r"uid=(\w+)", url)
    else:
        # NEF free look link format is:
        # https://ecf.almd.uscourts.gov/doc1/01713718205?caseid=75736&de_seq_num=30&magic_num=77910494
        match = re.search(r"magic_num=(\d+)", url)
    if match:
        return match.group(1)
    else:
        return None


def get_pacer_doc_id_from_doc1_url(url: str) -> str:
    """Extract the pacer document ID from the doc1 URL. Coerce the fourth digit
    to zero.

    In:  https://ecf.almd.uscourts.gov/doc1/01712427473
    Out: 01702427473
    In:  /doc1/01712427473
    Out: 01702427473

    Note that results are strings, not ints, because many of the strings start
    with zero.

    See tests for more examples.
    """
    assert (
        "show_case_doc" not in url
    ), "Cannot get doc1 ID from show_case_doc URL"
    url = url.rsplit("/", 1)[1].split("?")[0]
    url = f"{url[:3]}0{url[4:]}"
    return url


def get_pacer_seq_no_from_doc1_anchor(anchor):
    """Extract the PACER sequence number from an HTML anchor node.

    :param anchor: An LXML Element.
    :return: None if no sequence number can be found. Otherwise returns the
    sequence number.
    """
    try:
        onclick = anchor.xpath("./@onclick")[0]
    except IndexError:
        return None
    else:
        if "goDLS" in onclick:
            go_dls_parts = reverse_goDLS_function(onclick)
            return go_dls_parts["de_seq_num"]


def reverse_goDLS_function(s):
    """Extract the arguments from the goDLS JavaScript function.

    In: goDLS('/doc1/01712427473','56121','69','','','1','','');return(false);
    Out: {
      'form_post_url': '/doc1/01712427473',
      'caseid': '56121',
      'de_seq_num': '69',
      'got_receipt': '',
      'pdf_header': '',
      'pdf_toggle_possible': '1',
      'magic_num': '',
      'hdr': '',
    }

    The key names correspond to the form field names in the JavaScript on PACER,
    but we don't actually know what each of these values does. Our best
    speculation is:

     - form_post_url: Where the form is posted to. The HTML 'action' attribute.
     - caseid: The internal PACER ID for the case.
     - de_seq_num: Unclear. This seems to be the internal ID for the document,
       but this field can be omitted without any known issues.
     - got_receipt: If set to '1', this will bypass the receipt page and
       immediately direct you to the page where the PDF is embedded in an
       iframe.
     - pdf_header: Can be either 1 or 2. 1: Show the header. 2: No header.
     - pdf_toggle_possible: This seems to always be 1. Could be that some courts
       do not allow the header to be turned off, but we haven't discoered that
       yet.
     - magic_num: This is used for the "One free look" downloads.
     - hdr: Unclear what HDR stands for but on items that have attachments,
       passing this parameter bypasses the download attachment screen and takes
       you directly to the PDF that you're trying to download. For an example,
       see document 108 from 1:12-cv-00102 in tnmd, which is a free opinion that
       has an attachment. Note that the eighth parameter was added some time
       after 2010. Dockets older than that date only have seven responses.
    """
    args = re.findall("'(.*?)'", s)
    parts = {
        "form_post_url": args[0],
        "caseid": args[1],
        "de_seq_num": args[2],
        "got_receipt": args[3],
        "pdf_header": args[4],
        "pdf_toggle_possible": args[5],
        "magic_num": args[6],
    }
    try:
        parts["hdr"] = args[7]
    except IndexError:
        # At some point dockets added this eighth parameter. Older ones lack it
        parts["hdr"] = None
    return parts


def make_doc1_url(court_id, pacer_doc_id, skip_attachment_page):
    """Make a doc1 URL.

    If skip_attachment_page is True, we replace the fourth digit with a 1
    instead of a zero, which bypasses the attachment page.
    """
    if skip_attachment_page and pacer_doc_id[3] == "0":
        # If the fourth digit is a 0, replace it with a 1
        pacer_doc_id = f"{pacer_doc_id[:3]}1{pacer_doc_id[4:]}"
    doc_id_cid = get_court_id_from_doc_id_prefix(pacer_doc_id[:3])
    if court_id is None:
        court_id = doc_id_cid
    elif court_id != doc_id_cid:
        raise ValueError(
            f"pacer_doc_id {pacer_doc_id} prefix has court_id {doc_id_cid}, "
            f"expected {court_id}"
        )
    return f"https://ecf.{court_id}.uscourts.gov/doc1/{pacer_doc_id}"


def make_docs1_url(
    court_id: Optional[str], pacer_doc_id: str, skip_attachment_page
) -> str:
    """Make a docs1 URL for NDAs free look downloads.

    :param court_id: The court ID.
    :param pacer_doc_id: The PACER document ID.
    """
    if skip_attachment_page and pacer_doc_id[3] == "0":
        # If the fourth digit is a 0, replace it with a 1
        pacer_doc_id = f"{pacer_doc_id[:3]}1{pacer_doc_id[4:]}"
    doc_id_cid = get_court_id_from_doc_id_prefix(pacer_doc_id[:3])
    if court_id is None:
        court_id = doc_id_cid
    elif court_id != doc_id_cid:
        raise ValueError(
            f"pacer_doc_id {pacer_doc_id} prefix has court_id {doc_id_cid}, "
            f"expected {court_id}"
        )
    return f"https://ecf.{court_id}.uscourts.gov/docs1/{pacer_doc_id}"


def is_pdf(response):
    """Determines whether the item downloaded is a PDF or something else."""
    if response.headers.get("content-type") == "application/pdf":
        return True
    return False


def is_text(response):
    """Determines whether the item downloaded is a text file or something else."""
    if ".txt" in response.headers.get("content-type", ""):
        return True
    return False


def get_nonce_from_form(r):
    """Get a nonce value from a HTML response. Returns the first nonce that is
    found.

    :param r: The response object you wish to parse.
    :returns A nonce object that can be used to query PACER or None, if no
    nonce can be found.
    """
    try:
        tree = html.fromstring(r.text)
    except ValueError:
        # This usually happens when we are blocked from a court.
        raise ParsingException(
            "Got XML when expecting HTML and cannot parse it."
        )
    form_attrs = tree.xpath("//form//@action")
    for attr in form_attrs:
        # The action attr will be a value like:
        # ../cgi-bin/HistDocQry.pl?112801540788508-L_1_0-1
        # Split on the '?', and return the nonce.
        try:
            path, nonce = attr.split("?")
        except ValueError:
            raise ParsingException("Didn't get nonce from PACER form.")
        else:
            if "-L_" in nonce:
                return nonce
    return None


BASE_IA_URL = "https://www.archive.org/download"


def get_bucket_name(court, pacer_case_id):
    bucketlist = ["gov", "uscourts", court, str(pacer_case_id)]
    return ".".join(bucketlist)


def get_docket_filename(court, pacer_case_id):
    return ".".join(
        [
            "gov",
            "uscourts",
            str(court),
            str(pacer_case_id),
            "docket.xml",
        ]
    )


def get_document_filename(
    court, pacer_case_id, document_number, attachment_number
):
    return ".".join(
        [
            "gov",
            "uscourts",
            str(court),
            str(pacer_case_id),
            str(document_number),
            str(attachment_number or 0),
            "pdf",
        ]
    )


def get_docketxml_url(court, pacer_case_id):
    return "{}/{}/{}".format(
        BASE_IA_URL,
        get_bucket_name(court, pacer_case_id),
        get_docket_filename(court, pacer_case_id),
    )


def get_pdf_url(court, pacer_case_id, document_number, attachment_number):
    return "{}/{}/{}".format(
        BASE_IA_URL,
        get_bucket_name(court, pacer_case_id),
        get_document_filename(
            court, pacer_case_id, document_number, attachment_number
        ),
    )


def set_pacer_doc_id_as_appellate_document_number(
    de: Dict[str, Union[str, date, datetime]]
) -> None:
    """For appellate courts that don't use numbers, if available set the
    pacer_doc_id as document number.

    :param de: The docket entry dict to set the document number.
    :return: None, the dict is modified in place.
    """
    if not de["document_number"]:
        if de["pacer_doc_id"]:
            # If we lack the document number, but have
            # the pacer doc ID, use it.
            de["document_number"] = de["pacer_doc_id"]
        else:
            # We lack both the document number and the pacer doc ID.
            # Probably a minute order. No need to set either.
            pass


def parse_datetime_for_us_timezone(datetime_str: str) -> datetime:
    """Parse a datetime from a string that contains a US timezone.
    :param datetime_str: The str datetime to parse.
    :return: A datetime object with UTC timezone offset.
    """

    # Supported US timezones.
    tzinfos = {
        # US standard timezones
        "EST": gettz("US/Eastern"),
        "CST": gettz("US/Central"),
        "MST": gettz("US/Mountain"),
        "PST": gettz("US/Pacific"),
        "AKST": gettz("US/Alaska"),
        "HST": gettz("US/Hawaii"),
        "CHST": gettz("Pacific/Guam"),
        "SST": gettz("US/Samoa"),
        "AST": gettz("America/Puerto_Rico"),
        # US daylight saving time timezones
        "EDT": gettz("US/Eastern"),
        "CDT": gettz("US/Central"),
        "MDT": gettz("US/Mountain"),
        "PDT": gettz("US/Pacific"),
        "AKDT": gettz("US/Alaska"),
        "HDT": gettz("US/Hawaii"),
        # CHST, SST and AST, dont' observe DST.
    }
    date_time = parser.parse(datetime_str, tzinfos=tzinfos)
    if date_time.utcoffset() is None:
        # Raise an exception if a timezone abbreviation is not specified.
        raise NotImplementedError(f"Datetime {datetime_str} not understood.")
    return date_time



#  Scraper for the Court Of Chancery of Delaware
# CourtID: dechan
# Court Short Name: De.
# Author: Andrei Chelaru
# Reviewer: mlr
# Date created: 31 July 2014

from juriscraper.opinions.united_states.state import delaware


class Site(delaware.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "http://courts.delaware.gov/opinions/List.aspx?ag=Court%20of%20Chancery"
        self.court_id = self.__module__



"""
CourtID: illappct
Court Short Name: Ill. App. Ct.
Author: Rebecca Fordon
Reviewer: Mike Lissner
History:
* 2016-06-23: Created by Rebecca Fordon
* 2022-05-18: Updated by William E. Palin
"""

from juriscraper.oral_args.united_states.state import illappct_1st_dist


class Site(illappct_1st_dist.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.district = "WC"
        self._set_parameters()



"""
Scraper for Pennsylvania Supreme Court
CourtID: pa
Court Short Name: pa
"""
import re

from juriscraper.lib.string_utils import convert_date_string
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.regex = False
        self.url = "https://www.pacourts.us/Rss/Opinions/Supreme/"
        self.set_regex(r"(.*)(?:[,-]?\s+Nos?\.)(.*)")
        self.base = (
            "//item[not(contains(title/text(), 'Judgment List'))]"
            "[not(contains(title/text(), 'Reargument Table'))]"
            "[not(contains(title/text(), 'Order Amending Rules'))]"
            "[contains(title/text(), 'No.')]"
        )
        self.cases = []

    def _process_html(self):
        for item in self.html.xpath(self.base):
            creator = item.xpath("./creator/text()")[0]
            pubdate = item.xpath("./pubdate/text()")[0]
            title = item.xpath("./title/text()")[0]
            search = self.regex.search(title)
            url = item.xpath("./link/@href")[0]
            if search:
                name = search.group(1)
                docket = search.group(2)
            else:
                name = title
                docket = ""
            self.cases.append(
                {
                    "name": name,
                    "date": convert_date_string(pubdate),
                    "docket": docket,
                    "judge": creator,
                    "url": url,
                }
            )

    def _get_case_names(self):
        return [case["name"] for case in self.cases]

    def _get_download_urls(self):
        return [case["url"] for case in self.cases]

    def _get_case_dates(self):
        return [case["date"] for case in self.cases]

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.cases)

    def _get_docket_numbers(self):
        return [case["docket"] for case in self.cases]

    def _get_judges(self):
        return [case["judge"] for case in self.cases]

    def set_regex(self, pattern):
        self.regex = re.compile(pattern)



from juriscraper.opinions.united_states.state import mo


class Site(mo.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url_slug = "Eastern"
        self.url = self.build_url()

    def _get_divisions(self):
        return ["Eastern Dist."] * len(self.cases)



import pprint
import re
import sys

from ..lib.log_tools import make_default_logger
from ..lib.string_utils import force_unicode
from .reports import BaseReport
from .utils import (
    get_court_id_from_doc_id_prefix,
    get_pacer_doc_id_from_doc1_url,
    reverse_goDLS_function,
)

logger = make_default_logger()


class AttachmentPage(BaseReport):
    """An object for querying and parsing the attachment page report."""

    PATH = "doc1/"

    def __init__(self, court_id, pacer_session=None):
        super().__init__(court_id, pacer_session)
        # Note that parsing bankruptcy attachment pages does not reveal the
        # document number, only the attachment numbers.
        self.is_bankruptcy = self.court_id.endswith("b")

    def query(self, document_number):
        """Query the "attachment page" endpoint and set the results to self.response.

        :param document_number: The internal PACER document ID for the item.
        :return: a request response object
        """
        assert (
            self.session is not None
        ), "session attribute of DocketReport cannot be None."
        # coerce the fourth digit of the document number to 1 to ensure we get
        # the attachment page.
        document_number = f"{document_number[:3]}0{document_number[4:]}"
        url = self.url + document_number
        logger.info("Querying the attachment page endpoint at URL: %s", url)
        self.response = self.session.get(url)
        self.parse()

    @property
    def data(self):
        """Get data back from the query for the matching document entry.

        :return: If lookup fails, an empty dict. Else, a dict containing the
        following fields:
            - document_number: The document number we're working with.
            - page_count: The number of pages of the item
            - file_size_str: The size of the file as a str or null if unknown
            - pacer_doc_id: The doc ID for the main document.
            - attachments: A list of attached items with the following fields:
                - attachment_number: The attachment number.
                - description: A description of the item.
                - page_count: The number of pages.
                - file_size_str: The size of the file as a str or null if
                  unknown
                - pacer_doc_id: The document ID for the attachment (a str).

        See the JSON objects in the tests for more examples.
        """
        if self.is_valid is False:
            return {}

        rows = self.tree.xpath("//tr[.//a]")
        if not rows:
            logger.info("No documents found on attachment page.")
            return {}

        first_num = self._get_name_attachment_number(rows[0])
        # First row is an attachment
        if first_num is not None and first_num != 0:
            first_row = rows[0]
            first_row_attachment = True
        else:
            first_row = rows.pop(0)
            first_row_attachment = False
        result = {
            "document_number": self._get_document_number(),
            "pacer_doc_id": self._get_pacer_doc_id(first_row),
            "pacer_case_id": self._get_pacer_case_id(),
            "pacer_seq_no": self._get_pacer_seq_no_from_tr(first_row),
            "attachments": [],
        }
        if first_row_attachment:
            result["page_count"] = None
            result["file_size_str"] = ""
        else:
            result["page_count"] = self._get_page_count_from_tr(first_row)
            file_size_bytes = self._get_file_size_bytes_from_tr(first_row)
            if file_size_bytes is not None:
                result["file_size_bytes"] = file_size_bytes
            result["file_size_str"] = self._get_file_size_str_from_tr(
                first_row
            )
        for row in rows:
            attachment = {
                "attachment_number": self._get_attachment_number(row),
                "description": self._get_description_from_tr(row),
                "page_count": self._get_page_count_from_tr(row),
                "file_size_str": self._get_file_size_str_from_tr(row),
                "pacer_doc_id": self._get_pacer_doc_id(row),
                # It may not be needed to reparse the seq_no
                # for each row, but we may as well. So far, it
                # has always been the same as the main document.
                "pacer_seq_no": self._get_pacer_seq_no_from_tr(row),
            }
            file_size_bytes = self._get_file_size_bytes_from_tr(row)
            if file_size_bytes is not None:
                attachment["file_size_bytes"] = file_size_bytes
            result["attachments"].append(attachment)

        return result

    def _get_document_number(self):
        """Return the document number for an item.

        In district court attachment pages, this is easy to extract with an
        XPath. In bankruptcy cases, it's simply not there.
        """
        if self.is_bankruptcy:
            return None

        # First try inspecting the input elements
        input_els = self.tree.xpath("//input")
        for input_el in input_els:
            try:
                name = input_el.xpath("./@name")[0]
            except IndexError:
                continue
            else:
                split = name.split("_")
                # Document 16 name field example "document_16_0"
                if len(split) == 3 and split[0] == "document":
                    document_string = split[1]
                    # Any other matches will be invalid if this is empty
                    if document_string == "":
                        return None
                    document_number = int(document_string)
                    # Ensure document number is valid
                    if document_number != 0:
                        return document_number

        # There are two styles of attachment menus. Try them both.
        paths = (
            '//tr[contains(., "Document Number")]//a/text()',
            '//tr[contains(., "Main Document")]//a/text()',
        )
        for path in paths:
            try:
                return int(self.tree.xpath(path)[0].strip())
            except IndexError:
                continue
        return None

    def _decrement_attachment_index(self, row):
        """Return if we need to decrement the attachment index.

        We need to do this for all bankruptcy courts and old attachment pages
        in some district courts.

        For district courts we need to use a mapping table with the last dlsid
        that requires a decrement.

        Note that the decrement check must be done for each row in case any
        attachments in the page have sequences after the changeover as any
        updated attachments with a dlsid after the changeover should not be
        decremented.
        """
        sub_dlsid = {
            "akd": 682797,
            "ared": 1906030,
            "cacd": 9630188,
            "ctd": 2414673,
            "flmd": 7734601,
            "flnd": 2613796,
            "gamd": 995021,
            "gasd": 1159201,
            "hid": 1078891,
            "ilnd": 8278610,
            "iand": 899130,
            "innd": 1560410,
            "iasd": 1192972,
            "kyed": 2087957,
            "kywd": 1738484,
            "mad": 3618527,
            "mied": 4148337,
            "miwd": 2358030,
            "moed": 3299716,
            "msnd": 888190,
            "mssd": 2476698,
            "nced": 1868202,
            "ncmd": 1093535,
            "ndd": 299419,
            "ned": 1868173,
            "nvd": 3153010,
            "nyed": 6778644,
            "nynd": 1829022,
            "nysd": 8012889,
            "nywd": 1862410,
            "oked": 388419,
            "ord": 3322370,
            "pamd": 3122200,
            "sdd": 1239147,
            "txed": 3830781,
            "txnd": 4975453,
            "utd": 1745928,
            "vaed": 1338618,
            "waed": 1349144,
            "wawd": 3755107,
            "wied": 1420356,
            "wyd": 797428,
        }
        if self.is_bankruptcy:
            return True
        doc_id = self._get_pacer_doc_id(row)
        court_id = get_court_id_from_doc_id_prefix(doc_id[:3])
        if court_id not in sub_dlsid:
            return False
        dlsid = int(doc_id[3:])
        if sub_dlsid[court_id] < dlsid:
            return False
        return True

    def _get_name_attachment_number(self, row):
        try:
            name = row.xpath(".//input/@name")[0]
        except IndexError:
            return None
        else:
            split = name.split("_")
            # Document 16 name field example "document_16_0"
            if len(split) == 3 and split[0] == "document":
                return int(split[2])
        return None

    def _get_attachment_number(self, row):
        """Return the attachment number for an item.

        In district courts, this can be easily extracted. In bankruptcy courts,
        you must extract it, then subtract 1 from the value since these are
        tallied and include the main document.
        """
        number = self._get_name_attachment_number(row)
        if number is None:
            number = int(row.xpath(".//a/text()")[0].strip())
        if self._decrement_attachment_index(row):
            return number - 1
        return number

    def _get_description_from_tr(self, row):
        """Get the description from the row"""
        if not self.is_bankruptcy:
            index = 2
            # Some NEFs attachment pages for some courts have an extra column
            # (see nyed_123019137279), use index 3 to get the description
            columns_in_row = row.xpath(f"./td")
            if len(columns_in_row) == 5:
                index = 3
        else:
            index = 3

        description_text_nodes = row.xpath(f"./td[{index}]//text()")
        if not description_text_nodes:
            # No text in the cell.
            return ""
        description = description_text_nodes[0].strip()
        return force_unicode(description)

    @staticmethod
    def _get_input_value_from_tr(tr, idx):
        """Take a row from the attachment table and return the input value by
        index.
        """
        try:
            input = tr.xpath(".//input")[0]
        except IndexError:
            return None
        else:
            # initial value string "23515655-90555-2"
            # "90555" is size in bytes "2" is pages
            value = input.xpath("./@value")[0]
            split_value = value.split("-")
            if len(split_value) != 3:
                return None
            return split_value[idx]

    @staticmethod
    def _get_page_count_from_tr_input_value(tr):
        """Take a row from the attachment table and return the page count as an
        int extracted from the input value.
        """
        count = AttachmentPage._get_input_value_from_tr(tr, 2)
        if count is not None:
            return int(count)

    @staticmethod
    def _get_page_count_from_tr(tr):
        """Take a row from the attachment table and return the page count as an
        int extracted from the cell specified by index.
        """
        pg_cnt_input = AttachmentPage._get_page_count_from_tr_input_value(tr)
        if pg_cnt_input:
            return pg_cnt_input
        pg_cnt_str_nodes = tr.xpath('./td[contains(., "page")]/text()')
        if not pg_cnt_str_nodes:
            # It's a restricted document without page count information.
            return None

        for pg_cnt_str_node in pg_cnt_str_nodes:
            try:
                pg_cnt_str = pg_cnt_str_node.strip()
                return int(pg_cnt_str.split()[0])
            except ValueError:
                # Happens when the description field contains the
                # word "page" and gets caught by the xpath. Just
                # press on.
                continue

    @staticmethod
    def _get_file_size_bytes_from_tr(tr):
        """Take a row from the attachment table and return the number of bytes
        as an int.
        """
        file_size_str = AttachmentPage._get_input_value_from_tr(tr, 1)
        if file_size_str is None:
            return None
        file_size = int(file_size_str)
        if file_size == 0:
            return None
        return file_size

    @staticmethod
    def _get_file_size_str_from_tr(tr):
        """Take a row from the attachment table and return the number of bytes
        as an int.
        """
        cells = tr.xpath("./td")
        last_cell_contents = cells[-1].text_content()
        units = ["kb", "mb"]
        if any(unit in last_cell_contents.lower() for unit in units):
            return last_cell_contents.strip()
        return ""

    @staticmethod
    def _get_pacer_doc_id(row):
        """Take in a row from the attachment table and return the pacer_doc_id
        for the item in that row. Return None if the ID cannot be found.
        """
        try:
            url = row.xpath(".//a")[0]
        except IndexError:
            # Item exists, but cannot download document. Perhaps it's sealed
            # or otherwise unavailable in PACER. This is carried over from the
            # docket report and may not be needed here, but it's a good
            # precaution.
            return None
        else:
            doc1_url = url.xpath("./@href")[0]
            return get_pacer_doc_id_from_doc1_url(doc1_url)

    @staticmethod
    def _get_pacer_seq_no_from_tr(row):
        """Take a row of the attachment page, and return the sequence number
        from the goDLS function.
        """
        try:
            url = row.xpath(".//a")[0]
        except IndexError:
            # No link in the row. Maybe its sealed.
            pass
        else:
            try:
                onclick = url.xpath("./@onclick")[0]
            except IndexError:
                # No onclick on this row.
                pass
            else:
                if "goDLS" in onclick:
                    go_dls_parts = reverse_goDLS_function(onclick)
                    return go_dls_parts["de_seq_num"]

        # 1. Couldn't find a link in the row: Maybe it's sealed.
        # 2. No onclick on the row.
        # 3. No goDLS in the onclick
        return None

    def _get_pacer_case_id(self):
        """Get the pacer_case_id value by inspecting the HTML

        :returns str: The pacer_case_id value
        """
        # Start by inspecting all the links
        urls = self.tree.xpath("//a")
        for url in urls:
            try:
                onclick = url.xpath("./@onclick")[0]
            except IndexError:
                continue
            else:
                if "goDLS" not in onclick:
                    # Some other onclick we don't care about.
                    continue
                go_dls_parts = reverse_goDLS_function(onclick)
                return go_dls_parts["caseid"]

        # If that fails, try inspecting the input elements
        input_els = self.tree.xpath("//input")
        for input_el in input_els:
            try:
                onclick = input_el.xpath("./@onclick")[0]
            except IndexError:
                continue
            else:
                m = re.search(r"[?&]caseid=(\d+)", onclick, flags=re.I)
                if m:
                    return m.group(1)


def _main():
    if len(sys.argv) != 2:
        print("Usage: python -m juriscraper.pacer.attachment_page filepath")
        print("Please provide a path to an HTML file to parse.")
        sys.exit(1)
    report = AttachmentPage("cand")  # Court ID is only needed for querying.
    filepath = sys.argv[1]
    print(f"Parsing HTML file at {filepath}")
    with open(filepath) as f:
        text = f.read().decode("utf-8")
    report._parse_text(text)
    pprint.pprint(report.data, indent=2)


if __name__ == "__main__":
    _main()



"""Scraper for Dept of Justice Board of Immigration Appeals
CourtID: bia
Court Short Name: Dept of Justice Board of Immigration Appeals
Author: William Palin
Reviewer:
Type:
History:
    2021-12-18: Created by William E. Palin
"""
import re
from datetime import datetime
from typing import Any, Dict

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://www.justice.gov/eoir/ag-bia-decisions"
        self.article = None
        self.volume = 0
        self.urls = None

    def _process_elements(self, elements) -> Dict[str, Any]:
        """Process the element grouping.

        There is no easy way to parse our the content. Unfortunately, the DOJ
        admins randomly nest elements and not others. The only consistency
        is that they content is always split between HR tags.  So we iterate
        over the elements in order until we find an HR tag and then process
        the content.  Rinse Wash Repeat.

        Additionally, the only date we have is the year of the decision.

        :param elements: The elements between <hr> tags.
        :return: Case data
        """
        case = {}
        bold_text = elements[0].xpath(".//strong[1]/.. | .//b[1]/..")
        if not bold_text:
            return {}
        intro_text = (
            elements[0].xpath(".//strong[1]/.. | .//b[1]/..")[0].text_content()
        )
        intro_text = intro_text.replace(";", ",")
        name, cite = intro_text.split(",", 1)
        # Unfortunately there are no accessible file dates without PDF parsing
        # So we generate a date and mark it as date_filed_is_approximate = True
        # This is unset to false after it is extracted from the PDF on CL side.
        case["date_filed_is_approximate"] = True
        years = re.findall(r"\d{4}", cite)
        if not years:
            return {}
        case["date"] = f"{years[-1]}-07-01"
        case["status"] = "Published"
        case["citation"] = cite
        case["name"] = name
        case["url"] = elements[0].xpath(".//a")[0].get("href")
        case["docket"] = elements[0].xpath(".//a")[0].text_content()

        # Iterate over the P tags that hold the summaries, sometimes
        summary = []
        for element in elements:
            if element.tag == "p":
                summary.append(element.text_content())
        case["summary"] = "\n".join(summary).strip()
        return case

    def _process_html(self):
        if not self.test_mode_enabled():
            # Sort the URLS by volume to enable the backscraper
            # We reverse sort the links by volume and choose the first one
            # unless we are in a backscraper and then we choose what loop
            # we are in.
            if not self.urls:
                urls = self.html.xpath(
                    ".//table[1]/tbody/tr/td/a[contains(., 'Volume')]"
                )

                def get_text(elem):
                    return elem.text_content()

                self.urls = sorted(urls, key=get_text, reverse=True)
            self.url = self.urls[self.volume].get("href")
            # Download the new page of decisions
            self.html = super()._download()

        # Get the article which will contain all of our content.
        article = self.html.xpath(".//article")[0]
        # get the last element in the article
        # this ends the process_elements method on the final call because no
        # hr tag is present on the last decision

        last = list(article.iter())[-1]
        # Iterate over every tag in the article to separate out the cases.
        elements = []
        for element in article.iter():
            elements.append(element)
            # Process the data when the HR tag is found or the last element.
            # this loop lets us generate all of the elements and thus all
            # the data that we are looking for.  The DOJ has random and weird
            # HTML that sometimes nests and sometimes doesnt nest elements of
            # an opinion.
            if element.tag == "hr" or element == last:
                case = self._process_elements(elements)
                if case:
                    self.cases.append(case)
                elements, case = [], {}

    def extract_from_text(self, scraped_text: str) -> Dict[str, Any]:
        """Can we extract the date filed from the text?

        :param scraped_text: The content of the document downloaded
        :return: Metadata to be added to the case
        """
        date = re.findall(
            r"Decided (by (Acting\s)?Attorney General )?(.*\d{4})",
            scraped_text,
        )[0][-1]
        date_filed = datetime.strptime(date, "%B %d, %Y").strftime("%Y-%m-%d")
        metadata = {
            "OpinionCluster": {
                "date_filed": date_filed,
                "date_filed_is_approximate": False,
            },
        }
        return metadata



"""Scraper for US Coast Guard Court of Criminal Appeals
CourtID: cgcca
Court Short Name: C.G. Ct. Crim. App.

Author: Evando Blanco
Reviewer: flooie
History:
    2021-03-29: Created by Evando Blanco
    2021-12-17: Updated by flooie for OpinionSiteLinear
"""


import re
from typing import List

from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.citation_regex = (
            r"(?P<MJ>\(?\d{2} M\.J\. \d+\)?)|(?P<WL>\(?\d{4} (WL|Wl) \d+\)?)"
        )
        self.url = "https://www.uscg.mil/Resources/Legal/Court-of-Criminal-Appeals/CGCCA-Opinions/smdsort15701/publicationdate/smdorder15701/desc/"

    def _process_html(self) -> None:
        """Process the HTML

        :return: None
        """
        path = "//table[@class='Dashboard']/tbody/tr"
        for item in self.html.xpath(path):
            url = item.xpath(".//td/a/@href")[0]
            first_cell = item.xpath(".//td/a/text()")[0]
            docket = re.sub(
                r"Docket Nos?\.", "", item.xpath(".//td[2]/text()")[0]
            )
            date_string = item.xpath(".//td[3]/text()")[0]
            m = re.search(self.citation_regex, first_cell)
            if m:
                mj = m.group("MJ").strip("()") if m.group("MJ") else ""
                wl = m.group("WL").strip("()") if m.group("WL") else ""
            else:
                mj = ""
                wl = ""

            status = "Published" if m and mj else "Unpublished"

            self.cases.append(
                {
                    "name": first_cell,
                    "date": date_string,
                    "docket": docket,
                    "url": url,
                    "status": status,
                    "citation": mj,
                    "parallel_citation": wl,
                }
            )

    def _get_case_names(self) -> List[str]:
        """Clean case names

        :return: List of case names
        """
        for case in self.cases:
            if case["citation"]:
                case["name"] = re.sub(
                    rf"\(?\s?{case['citation']}\s?\)?",
                    "",
                    case["name"],
                )
            if case["parallel_citation"]:
                case["name"] = re.sub(
                    rf"\(?\s?{case['parallel_citation']}\s?\)?",
                    "",
                    case["name"],
                )
            case["name"] = re.sub(
                r"\(UNPUBLISHED\)|\(MERITS\)|\(PER CURIAM\)|ORDER",
                "",
                case["name"],
                flags=re.IGNORECASE,
            )
            case["name"] = titlecase(case["name"].strip(" -"))

        return [c["name"] for c in self.cases]



import time
from datetime import date

from lxml import html

from juriscraper.opinions.united_states_backscrapers.federal_special import (
    cit_2010,
)


class Site(cit_2010.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "http://www.cit.uscourts.gov/SlipOpinions/SlipOps-2009.html"
        self.court_id = self.__module__

    def _get_case_dates(self):
        case_dates = []
        for e in self.html.xpath("//table[3]/tr/td[3][../td/a]"):
            date_string = html.tostring(
                e, method="text", encoding="unicode"
            ).strip()
            if date_string == "05/13/20009":
                # Special case.
                date_string = "05/13/2009"
            case_dates.append(
                date.fromtimestamp(
                    time.mktime(time.strptime(date_string, "%m/%d/%Y"))
                )
            )
        return case_dates

    def _get_nature_of_suit(self):
        return None



import random
import time

from requests.adapters import HTTPAdapter
from requests.packages.urllib3.poolmanager import PoolManager

from juriscraper.AbstractSite import logger


class SSLAdapter(HTTPAdapter):
    """An HTTPS Transport Adapter that uses an arbitrary SSL version."""

    def __init__(self, ssl_version=None, **kwargs):
        self.ssl_version = ssl_version

        super().__init__(**kwargs)

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = PoolManager(
            num_pools=connections,
            maxsize=maxsize,
            block=block,
            ssl_version=self.ssl_version,
        )


def add_delay(delay=0, deviation=0):
    """Create a semi-random delay.

    Delay is the number of seconds your program will be stopped for, and
    deviation is the number of seconds that the delay can vary.
    """
    duration = random.randrange(delay - deviation, delay + deviation)
    logger.info(f"Adding a delay of {duration} seconds. Please wait.")
    time.sleep(duration)



# Author: Michael Lissner
# Date created: 2013-05-23


from juriscraper.opinions.united_states.state import haw


class Site(haw.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.target_id = "ICA"



# Auth: arderyp
# Date: 2017-01-26

from datetime import date

from lxml import html

from juriscraper.opinions.united_states.state import ohioctcl


class Site(ohioctcl.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__

        # If it's January, we need to make sure that we aren't
        # missing any late-coming cases from the previous year.
        today = date.today()
        if today.month == 1:
            self.year = str(today.year - 1)
        else:
            # This simply aborts the crawler.
            self.status = 200
            self.html = html.fromstring("<html></html>")



import time
from datetime import date

from lxml import html

from juriscraper.opinions.united_states.federal_special import cit


class Site(cit.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "http://www.cit.uscourts.gov/SlipOpinions/SlipOps-2010.html"
        self.court_id = self.__module__

    def _get_case_dates(self):
        case_dates = []
        for e in self.html.xpath("//table[3]/tr/td[3][../td/a]"):
            date_string = html.tostring(
                e, method="text", encoding="unicode"
            ).strip()
            if date_string == "06/25//2010":
                # Special case.
                date_string = "06/25/2010"
            case_dates.append(
                date.fromtimestamp(
                    time.mktime(time.strptime(date_string, "%m/%d/%Y"))
                )
            )
        return case_dates

    def _get_download_urls(self):
        return [t for t in self.html.xpath("//table[3]//tr/td[1]/a/@href")]

    def _get_citations(self):
        return [t for t in self.html.xpath("//table[3]//tr/td[1]/a/text()")]

    def _get_case_names(self):
        # Exclude confidential rows
        case_names = []
        for e in self.html.xpath("//table[3]/tr[position() > 1]/td[2]"):
            s = html.tostring(e, method="text", encoding="unicode")
            if "confidential" in s:
                continue
            elif "Errata" in s:
                index = s.find("Errata")
            elif "Public version" in s:
                index = s.find("Public version")
            else:
                index = -1
            case_names.append(s[:index] if index > -1 else s)
        return case_names

    def _get_precedential_statuses(self):
        statuses = []
        for e in self.html.xpath(
            "//table[3]//tr[position() > 1]/td[2][../td/a]"
        ):
            s = (
                html.tostring(e, method="text", encoding="unicode")
                .lower()
                .strip()
            )
            if "errata" in s:
                statuses.append("Errata")
            else:
                statuses.append("Published")
        return statuses

    def _get_docket_numbers(self):
        docket_numbers = []
        for e in self.html.xpath(
            "//table[3]//tr[position() > 1]/td[4][../td/a]"
        ):
            docket_numbers.append(
                html.tostring(e, method="text", encoding="unicode").strip()
            )
        return docket_numbers

    def _get_judges(self):
        judges = []
        for e in self.html.xpath(
            "//table[3]//tr[position() > 1]/td[5][../td/a]"
        ):
            s = html.tostring(e, method="text", encoding="unicode")
            judges.append(s)
        return judges

    def _get_nature_of_suit(self):
        return [
            t for t in self.html.xpath("//table[3]//tr/td[6][../td/a]/text()")
        ]



# Scraper for Florida 2nd District Court of Appeal Per Curiam
# CourtID: flaapp2
# Court Short Name: flaapp2

from . import fladistctapp_1_per_curiam


class Site(fladistctapp_1_per_curiam.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court_number = 2
        self.url = self.get_url()
        self.back_scrape_iterable = [
            (11, 2017),
            (12, 2017),
            (1, 2018),
            (2, 2018),
        ]



# Scraper for New York Appellate Divisions 2nd Dept.
# CourtID: nyappdiv_2nd
# Court Short Name: NY
# Author: Andrei Chelaru
# Reviewer: mlr
# Date: 2014-07-04

from juriscraper.opinions.united_states.state import nyappdiv_1st


class Site(nyappdiv_1st.Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.division = 2
        self.url = self.build_url()



"""Scraper for United States Court of Appeals for the Federal Circuit
CourtID: cafc
Court Short Name: U.S. Court of Appeals for the Federal Circuit
Author: Dave Voutila
Reviewer:
Type:
History:
    2016-05-10: Created by Dave Voutila, @voutilad
    2021-01-24: Item's title split fixed to extract opinions name, @satsuki-chan
"""

import feedparser
from lxml.html import fromstring

from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "https://cafc.uscourts.gov/category/opinion-order/feed/"
        self.court_id = self.__module__

    def _process_html(self) -> None:
        """Process the RSS feed.

        Iterate over each item in the RSS feed to extract out
        the date, case name, docket number, and status and pdf URL.
        Return: None
        """
        feed = feedparser.parse(self.request["response"].content)
        for item in feed["entries"]:
            value = item["content"][0]["value"]
            docket, title = item["title"].split(" [")[0].split(": ", 1)
            slug = fromstring(value).xpath(".//a/@href")[0]
            self.cases.append(
                {
                    "date": item["published"],
                    "docket": docket,
                    "url": f"https://cafc.uscourts.gov{slug}",
                    "name": titlecase(title),
                    "status": self._get_status(item["title"].lower()),
                }
            )

    def _get_status(self, title: str) -> str:
        """Get precedential status from title string.

        return: The precedential status of the case.
        """
        if "nonprecedential" in title:
            status = "Unpublished"
        elif "precedential" in title:
            status = "Published"
        else:
            status = "Unknown"
        return status



"""Scraper for Louisiana Court of Appeal, First Circuit
CourtID: lactapp_1
Court Short Name: La. Ct. App. 1st Cir.
Author: mmantel
History:
  2019-11-24: Created by mmantel
"""

import math
import re

from juriscraper.lib.html_utils import (
    get_row_column_links,
    get_row_column_text,
)
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self._page_size = 50
        self._base_url = f"https://www.la-fcca.org/opiniongrid/opinionpub.php?opinionpage_size={self._page_size}"
        self.url = self._base_url
        self.back_scrape_iterable = self._generate_back_scrape_range()

        # The opinions page does not indicate whether a case is
        # published or unpublished. That is only found in the PDF.
        # (Unpublished cases have "Not Designated For Publication" on
        # the cover page.)
        self.status = "Unknown"

    def _process_html(self):
        for row in self.html.cssselect("#opinion_contentTable tbody tr"):
            self.cases.append(
                {
                    "date": get_row_column_text(row, 1),
                    "docket": self._parse_docket_numbers(row),
                    "name": get_row_column_text(row, 4),
                    "url": get_row_column_links(row, 3),
                }
            )

    def _parse_docket_numbers(self, row):
        # Handle cases such as:
        # "2018CA1765<br>Rehearing<br>Application"
        #     => "2018CA1765"
        # "2018CA1742<br>Consolidated With<br>2018CA1743"
        #     => "2018CA1742, 2018CA1743"
        text = get_row_column_text(row, 2)
        case_numbers = re.findall("[0-9]{4}[A-Z]{2}[0-9]{4}", text)
        return ", ".join(case_numbers)

    def _generate_back_scrape_range(self):
        # This is a generator function, so this code won't run until a
        # caller begins iterating, which is necessary because
        # otherwise this would run during unit tests and trigger an
        # unwanted network request.
        last_page = self._get_last_page_number()

        yield from range(1, last_page + 1)

    def _get_last_page_number(self):
        # The link to the last page has an onclick like:
        # javascript:opinion_doPostBack('paging','','&opinionsort_field=sortdate&opinionsort_field_by=&opinionsort_field_type=&opinionsort_type=DESC&opinionpage_size=50&opinionp=395')
        # where 395 is the last page number.
        html = self._get_html_tree_by_url(self._base_url, {})
        el = html.cssselect("a[title=last]")[0]
        onclick = el.get("onclick")
        return int(re.findall(r"\d+", onclick)[-1])

    def _download_backwards(self, page):
        self.url = self._base_url + ("&opinionp=%d" % page)
        self.html = self._download()
        self._process_html()



from juriscraper.OpinionSiteLinear import OpinionSiteLinear
from juriscraper.WebDriven import WebDriven


class OpinionSiteLinearWebDriven(OpinionSiteLinear, WebDriven):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        WebDriven.__init__(self, args, kwargs)

    def __del__(self):
        self.close_session()
        self.close_webdriver_session()



# Scraper for Georgia Appeals Court
# CourtID: gactapp
# Court Short Name: gactapp
# Author: Andrei Chelaru
# Reviewer: mlr
# Date created: 25 July 2014


from datetime import date, timedelta

from juriscraper.lib.string_utils import convert_date_string, titlecase
from juriscraper.OpinionSite import OpinionSite


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.case_date = date.today()
        self.a_while_ago = date.today() - timedelta(days=28)
        self.base_path = "id('art-main')//tr[position() > 1]"
        self.url = (
            f"http://www.gaappeals.us/docketdate/results_all.php?searchterm="
            f"{self.a_while_ago.month:02d}%2F{self.a_while_ago.day:02d}%2F{self.a_while_ago.year}&searchterm2="
            f"{self.case_date.month:02d}%2F{self.case_date.day:02d}%2F{self.case_date.year}&submit=Search"
        )

    def _get_case_names(self):
        path = f"{self.base_path}/td[2]/text()"
        return [titlecase(e) for e in self.html.xpath(path)]

    def _get_download_urls(self):
        path = f"{self.base_path}/td[6]/a/@href"
        return list(self.html.xpath(path))

    def _get_case_dates(self):
        path = f"{self.base_path}/td[3]/text()"
        return [convert_date_string(e) for e in self.html.xpath(path)]

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)

    def _get_docket_numbers(self):
        path = f"{self.base_path}/td[1]/text()"
        return list(self.html.xpath(path))

    def _get_dispositions(self):
        path = f"{self.base_path}/td[4]/text()"
        return [titlecase(e) for e in self.html.xpath(path)]



# Scraper for Kansas Appeals Court
# CourtID: kan_p

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://www.kscourts.org/cases-opinions/opinions.html"
        self.status = "Published" if "_p" in self.court_id else "Unpublished"
        self.xpath = "kc-scu" if "kan_" in self.court_id else "kc-cau"
        self.rows = "preceding" if "_p" in self.court_id else "following"

    def _process_html(self):
        h2_element = self.html.xpath(f'//h2[@id="{self.xpath}"]')[0]
        ul_elements = h2_element.xpath(f"{self.rows}-sibling::ul")
        for ul in ul_elements:
            if "None Released" in ul.text_content():
                continue
            date = ul.getprevious().text_content()[6:]
            for item in ul.xpath(".//p"):
                components = item.text_content().split(" - ")
                self.cases.append(
                    {
                        "status": self.status,
                        "date": date,
                        "docket": components[0],
                        "name": components[1],
                        "url": item.xpath(".//a/@href")[0],
                    }
                )



"""
CourtID: illappct
Court Short Name: Ill. App. Ct.
Author: Rebecca Fordon
Reviewer: Mike Lissner
History:
* 2016-06-23: Created by Rebecca Fordon
* 2022-05-18: Updated by William E. Palin
"""
from datetime import timedelta

from dateutil.utils import today

from juriscraper.OralArgumentSiteLinear import OralArgumentSiteLinear


class Site(OralArgumentSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://www.illinoiscourts.gov/courts/appellate-court/oral-argument-audio/"
        self.method = "POST"
        self.today = today().strftime("%m/%d/%Y")
        self.earlier = (today() - timedelta(days=30)).strftime("%m/%d/%Y")
        self.district = "1st"
        self._set_parameters()

    def _set_parameters(self):
        self.parameters = {
            "__EVENTTARGET": "ctl00$ctl04$ddlFilterDistrict",
            "__EVENTARGUMENT": "",
            "__LASTFOCUS": "",
            "__VIEWSTATEGENERATOR": "66677877",
            "ctl00$header$search$txtSearch": "",
            "ctl00$ctl04$txtFilterName": "",
            "ctl00$ctl04$ddlFilterDate": "Custom Date Range",
            "ctl00$ctl04$txtFilterDateFrom": self.earlier,
            "ctl00$ctl04$txtFilterDateTo": self.today,
            "ctl00$ctl04$ddlFilterDistrict": self.district,
            "ctl00$ctl04$hdnSortField": "ArgumentDate",
            "ctl00$ctl04$hdnSortDirection": "DESC",
        }

    def _process_html(self):
        for row in self.html.xpath('.//table[@id="ctl04_gvArguments"]/tr')[2:]:
            url = row.xpath(".//a/@data-audio")[0].replace(" ", "%20")
            date, district, docket, name, *others = row.xpath(".//span/text()")
            if others:
                docket, name = name, others[0]
            self.cases.append(
                {
                    "date": date,
                    "docket": docket,
                    "name": name,
                    "district": district,
                    "url": url,
                }
            )



from typing import Optional

import requests
from requests import Response

from juriscraper.pacer.reports import BaseReport

from ..lib.log_tools import make_default_logger
from .docket_report import BaseDocketReport

logger = make_default_logger()


class ListOfCreditors(BaseDocketReport, BaseReport):
    """Query and parse the List of Creditors report."""

    FORMAT_RAW_DATA_SERVICE = "https://ncrs.uscourts.gov/query/cmecf/index.adp"
    PATH = "cgi-bin/CredMatrixCase.pl"
    CACHE_ATTRS = ["metadata"]
    ERROR_STRINGS = BaseReport.ERROR_STRINGS + [
        "You entered a non-bankruptcy type case:",
    ]

    def __init__(self, court_id, pacer_session=None):
        BaseDocketReport.__init__(self, court_id)
        BaseReport.__init__(self, court_id, pacer_session)

        self._clear_caches()
        self._metadata = None
        assert court_id.endswith(
            "b"
        ), "Unable to create object. Must use bankruptcy court abbreviation."

    @property
    def data(self):
        """Get all the data back from this endpoint."""
        if self.is_valid is False:
            return {}

        data = self.metadata.copy()
        return data

    @property
    def metadata(self):
        """Parse the raw data from the HTML file."""
        if self.is_valid is False:
            return {}

        if self._metadata is not None:
            return self._metadata

        try:
            raw_data = self.tree.xpath(
                '//form[@name="bnc"]//input[@name="data"]/@value'
            )[0]
        except IndexError:
            raw_data = None
        meta_data = {
            "court_id": self.court_id,
            "data": raw_data,
        }
        self._metadata = meta_data
        return meta_data

    def _get_valid_post_param(self) -> str:
        """Get a valid POST param from the input form.
        From: /cgi-bin/CredMatrixCase.pl?209635477353597-L_1_0-1
            "209635477353597-L_1_0-1"
        return: A valid POST param to query the report.
        """

        action_value = self.tree.xpath('//form[@method="POST"]/@action')[0]
        param = action_value.split("?")[1]
        return param

    def download_file(self) -> Optional[Response]:
        """Downloads the formated pipe-limited file using the
        FORMAT_RAW_DATA_SERVICE API.

        :return: The server's response if the file content is valid, or None if
        an error occurred.
        """

        params = {
            "format": "rawdata",
            "useragentstring": "CM/ECF-BK V10.6.4",
            "data": self._metadata["data"],
        }
        req_timeout = (60, 300)
        r = requests.post(
            self.FORMAT_RAW_DATA_SERVICE, params=params, timeout=req_timeout
        )
        if "text/plain" in r.headers.get("content-type"):
            return r
        return None

    def query_post_param(self) -> str:
        """To query the report and ensure that the cost is the same as in the
        browser, obtain a valid POST param 'x-L_1_0-1' from the input form.

        :return: A valid query POST param 'x-L_1_0-1'
        """
        assert (
            self.session is not None
        ), "session attribute of ListOfCreditors cannot be None."

        logger.info(f"Getting a valid POST param for '{self.court_id}'")
        self.response = self.session.get(self.url)
        self.parse()
        return self._get_valid_post_param()

    def query(
        self,
        pacer_case_id: str,
        docket_number: str,
        post_param: str,
    ):
        """Query the List of creditors report and return the results.

        :param pacer_case_id: The internal PACER ID for the case.
        :param docket_number: A docket number to look up. Something like
        2:17-bk-39239.

        :return: request response object
        """
        assert (
            self.session is not None
        ), "session attribute of ListOfCreditors cannot be None."

        params = {
            "all_case_ids": pacer_case_id,
            "case_num": docket_number,
            "typfmt": "rawfrmt",  # Returns raw data.
            "SMG": "",  # Special mailing group empty.
        }

        logger.info(
            "Querying list of creditors for case ID '%s' in court '%s' "
            "with params %s",
            pacer_case_id,
            self.court_id,
            params,
        )
        self.response = self.session.post(
            f"{self.url}?{post_param}", data=params
        )
        self.parse()



# Scraper for New York Appellate Term 1st Dept.
# CourtID: nyappterm_1st
# Court Short Name: NY

import re
from datetime import date, timedelta
from typing import Any, Dict

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court = "Appellate Term, 1st Dept"
        self.court_id = self.__module__
        self.url = "https://iapps.courts.state.ny.us/lawReporting/Search?searchType=opinion"
        self._set_parameters()

    def _set_parameters(self) -> None:
        """Set the parameters for the POST request."""
        self.method = "POST"
        today = date.today()
        self.parameters = {
            "rbOpinionMotion": "opinion",
            "Pty": "",
            "and_or": "and",
            "dtStartDate": (today - timedelta(days=30)).strftime("%m/%d/%Y"),
            "dtEndDate": today.strftime("%m/%d/%Y"),
            "court": self.court,
            "docket": "",
            "judge": "",
            "slipYear": "",
            "slipNo": "",
            "OffVol": "",
            "Rptr": "",
            "OffPage": "",
            "fullText": "",
            "and_or2": "and",
            "Order_By": "Party Name",
            "Submit": "Find",
            "hidden1": "",
            "hidden2": "",
        }

    def _process_html(self):
        for row in self.html.xpath(".//table")[-1].xpath(".//tr")[1:]:
            slip_cite = " ".join(row.xpath("./td[5]//text()"))
            official_citation = " ".join(row.xpath("./td[4]//text()"))
            url = row.xpath(".//a")[0].get("href")
            url = re.findall(r"(http.*htm)", url)[0]
            status = "Unpublished" if "(U)" in slip_cite else "Published"
            self.cases.append(
                {
                    "name": row.xpath(".//td")[0].text_content(),
                    "date": row.xpath(".//td")[1].text_content(),
                    "url": url,
                    "status": status,
                    "docket": "",
                    "citation": official_citation,
                    "parallel_citation": slip_cite,
                }
            )

    def extract_from_text(self, scraped_text: str) -> Dict[str, Any]:
        """Can we extract the date filed from the text?

        :param scraped_text: The content of the document downloaded
        :return: Metadata to be added to the case
        """
        # Ny App Term 1st Dept. 2nd and Sup Ct all have different varying
        # docket number types.
        # ie. 123413/03 vs. 51706 vs. 2003-718 Q C or 2003-1288 K C

        dockets = re.findall(
            r"(\d+\/\d+)|^(\d{5,})|^(\d+-\d+ \w+\s\w+)", scraped_text
        )
        dockets = [list(filter(None, x)) for x in dockets]
        metadata = {
            "Docket": {
                "docket_number": dockets[0][0] if dockets else "",
            },
        }
        return metadata



# Author: Michael Lissner
# Date created: 2013-06-03
# Date updated: 2020-02-25

from juriscraper.lib.string_utils import convert_date_string
from juriscraper.OpinionSite import OpinionSite


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://juddocumentservice.mt.gov/getDailyOrders"
        self.download_base = (
            "https://juddocumentservice.mt.gov/getDocByCTrackId?DocId="
        )

    def _get_download_urls(self):
        return [f"{self.download_base}{row['cTrackId']}" for row in self.html]

    def _get_case_names(self):
        return [f"{row['title']}" for row in self.html]

    def _get_case_dates(self):
        return [convert_date_string(f"{row['fileDate']}") for row in self.html]

    def _get_precedential_statuses(self):
        return [
            "Published"
            if "Published" in row["documentDescription"]
            else "Unpublished"
            for row in self.html
        ]

    def _get_docket_numbers(self):
        return [f"{row['caseNumber']}" for row in self.html]

    def _get_summaries(self):
        return [f"{row['documentDescription']}" for row in self.html]

    def _get_nature_of_suit(self):
        natures = []

        for docket in self.docket_numbers:
            if docket.startswith("DA"):
                nature = "Direct Appeal"
            elif docket.startswith("OP"):
                nature = "Original Proceeding"
            elif docket.startswith("PR"):
                nature = "Professional Regulation"
            elif docket.startswith("AF"):
                nature = "Administrative File"
            else:
                nature = "Unknown"
            natures.append(nature)
        return natures



"""
Scraper for New York Court of Appeals
CourtID: ny
Court Short Name: NY
History:
 2015-10-27  Created by Andrei Chelaru
"""
from datetime import date

from dateutil.rrule import DAILY, rrule

from juriscraper.opinions.united_states.state.nyappterm_1st import (
    Site as NySite,
)


class Site(NySite):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.interval = 200
        self.back_scrape_iterable = [
            i.date()
            for i in rrule(
                DAILY,
                interval=self.interval,
                dtstart=date(2003, 6, 1),
                until=date(2016, 1, 1),
            )
        ]
        self.court = "Court of Appeals"



"""
Scraper for New Hampshire Supreme Court
CourtID: nh
Court Short Name: NH
Court Contact: webmaster@courts.state.nh.us
Author: Andrei Chelaru
Reviewer: mlr
History:
    - 2014-06-27: Created
    - 2014-10-17: Updated by mlr to fix regex error.
    - 2015-06-04: Updated by bwc so regex catches comma, period, or
    whitespaces as separator. Simplified by mlr to make regexes more semantic.
    - 2016-02-20: Updated by arderyp to handle strange format where multiple
    case names and docket numbers appear in anchor text for a single case pdf
    link. Multiple case names are concatenated, and docket numbers are
    concatenated with ',' delimiter
    - 2021-12-29: Updated for new web site, by flooie and satsuki-chan
Notes:
    To filer documents per year, the API expects the key number for the year
    in parameter:
    - 'tag': for years 2019 to date
        * 2021: tag=1206    * 2020: tag=1366    * 2019: tag=1416
    - 'subcategory': for years 2009 to 2018
        * 2018: subcategory=1601    * 2017: subcategory=1596
        * 2016: subcategory=1591    * 2015: subcategory=1586
        * 2014: subcategory=1581    * 2013: subcategory=1576
        * 2012: subcategory=1571    * 2011: subcategory=1566
        * 2010: subcategory=1561    * 2009: subcategory=1556
    Only one parameter can have a key number; using both with a value in the
    same request (even with a valid key) returns zero documents.
    Using none, returns all documents (regardless of the publication year)
    found with the other parameters values.
    For this scraper, filtering documents per year is not necesary.
    Examples:
        - https://www.courts.nh.gov/content/api/documents?text=&category=+undefined&purpose=1331+undefined&tag=+&subcategory=1556&sort=field_date_posted|desc&page=3&size=10
        - https://www.courts.nh.gov/content/api/documents?text=&category=+undefined&purpose=1331+undefined&tag=1366+&subcategory=&sort=field_date_posted|desc&page=1&size=25
"""

import re

from lxml.html import fromstring

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        # Parameters are included in the URL because this court's API only
        # allows GET requests. Current architecture for scrapers only includes
        # parameters in requests with the POST method
        self.url = "https://www.courts.nh.gov/content/api/documents?sort=field_date_posted%7Cdesc&page=1&size=30&purpose=1331"
        self.status = "Published"

    def _process_html(self) -> None:
        for case in self.html["data"]:
            content = fromstring(case["documentContent"])
            urls = content.xpath(
                ".//div[@class='document__detail__title']//a/@href"
            )
            if not urls:
                logger.info(f"Opinion without URL to file: {case}")
                continue

            date = case["documentPosted"]
            if not date:
                logger.info(f"Opinion without date: {case}")
                continue

            # In juvenile cases - the case name contains the docket number
            # So we use a negative lookbehind to ignore docket numbers after
            # the word juvenile.
            regex = r"(?<!JUVENILE )((\b)(\w+-)?[\d]{2,4}-\d+)\/?"
            dockets = re.findall(regex, case["documentName"])
            docket = ", ".join([docket[0] for docket in dockets])

            # Get name by removing docket number from document text
            rgx = r"(?<!JUVENILE )((\b)(\w+-)?[\d]{2,4}-\d+)\/?,? &? ?(and)? ?"
            name = re.sub(rgx, "", case["documentName"])
            # Strip out the bad pattern with slashes between adjoined docket numbers
            name = re.sub(r"^\d+\/", "", name)

            self.cases.append(
                {
                    "name": name,
                    "docket": docket,
                    "date": date,
                    "url": urls[0],
                }
            )
